from graphviz import Digraph


class CompGraphVisualizer:
    """
        CompGraphVisualizer is the building block object for quick 
        visualization of a computational graph. Nanograd dynamically 
        builds a computational graph and keeps track of the operations 
        made while building the computational graph.

        Performing DFS over the computational graph, CompGraphVisualizer
        is able to build a trace from a root node. Usually the root node
        is simply the final result of an operation. For a MLP, it usually
        is the loss function.

        ..note: CompGraphVisualizer is made to visualize a computational 
        graph. To have a more holistic view of a neural network architecture,
        prefer the use of NetworkVisualizer. 
    """
    def __init__(self):
        self.nodes, self.edges = set(), set()
    
    def visualize(self, root, rankdir="LR"):
        """
            Builds the computational graph and displays it.

            Args:
                root (Tensor): node to start building the trace.
                rankdir (str): graph organization
        """
        self._build_trace(root)
        graph = self._build_graph(rankdir=rankdir)
        return graph
    
    def _build_trace(self, node):
        """Performs a recursive depth-first search over the computational graph."""
        if node not in self.nodes:
            self.nodes.add(node)
            for child in node.children:
                self.edges.add((child, node))
                self._build_trace(child)
    
    def _build_graph(self):
        raise NotImplementedError("build_graph() must be implemented for subclasses!")


class ForwardGraphVisualizer(CompGraphVisualizer):
    def __init__(self):
        super().__init__()
    
    def _build_graph(self, rankdir:str='LR'):
        r"""
            Plots the forward graph

            Args:
                rankdir (str): TB (top to bottom graph) | LR (left to right)

            ..note: A node is uniquely identified with the id() function which guarantees
                    a unique number for every Python object.
        """
        assert rankdir in ['LR', 'TB'], f"Unexpected rankdir argument (TB, LR available). Got {rankdir}."
        graph = Digraph(format='png', graph_attr={'rankdir': rankdir})
        
        for n in self.nodes:
            name = n.name if n.name != "no_name" else (n.op + '_res' if n.op else n.name)
            graph.node(name=str(id(n)), label = f"{name} | {n.shape}", shape='record')
            if n.op:
                graph.node(name=str(id(n)) + n.op, label=n.op)
                graph.edge(str(id(n)) + n.op, str(id(n)))
        
        for n1, n2 in self.edges:
            if n2.op:
                graph.edge(str(id(n1)), str(id(n2)) + n2.op)
            else:
                graph.edge(str(id(n1)), str(id(n2)))
        
        return graph


class BackwardGraphVisualizer(CompGraphVisualizer):
    def __init__(self):
        super().__init__()
    
    def _build_graph(self, rankdir:str='LR'):
        r"""  
            Plots the backward graph

            Args:
                rankdir (str): TB (top to bottom graph) | LR (left to right)

            ..note: A node is uniquely identified with the id() function which guarantees
                    a unique number for every Python object.
        """
        assert rankdir in ['LR', 'TB'], f"Unexpected rankdir argument (TB, LR available). Got {rankdir}."
        graph = Digraph(format='png', graph_attr={'rankdir': rankdir})
        
        for n in self.nodes:
            name = n.name if n.name != "no_name" else (n.op + '_res' if n.op else n.name)
            graph.node(
                name=str(id(n)), 
                label=f"{name} | {n.shape} | {n.grad_fn.function_name}" if n.grad_fn else f"{name} | {n.shape}",
                shape="record")
            
        for n1, n2 in self.edges:
            graph.edge(str(id(n1)), str(id(n2)))
        
        return graph
        



# Inspired from: https://github.com/geohot/tinygrad/blob/master/tinygrad/ops_gpu.py
import numpy as np
import functools

from nanograd.nn.conv_ops import get_conv1d_output_size, get_conv2d_output_size

try:
    import pyopencl as cl
    PYOPENCL_AVAILABLE = True

except ImportError:
    PYOPENCL_AVAILABLE = False
    warnings.warn("PyOpenCL is not available on this computer. Can't use \
                   parallel computing. Please install it to move comptutations \
                   to move the GPU.")

# *************************************
# ****** OpenCL helper functions ******
# *************************************

class GPUBuffer:
    def __init__(self, ctx, shape, hostbuf=None):
        if isinstance(shape, int): shape = [shape]
        self.shape, self.dtype = tuple(shape), np.float32

        if isinstance(hostbuf, GPUBuffer):
            self.cl = hostbuf.cl
        else:
            self.cl = cl.Buffer(ctx, cl.mem_flags.READ_WRITE | (cl.mem_flags.COPY_HOST_PTR if hostbuf is not None else 0), 
                                np.int32(4*np.prod(shape)), hostbuf=hostbuf.astype(np.float32).ravel() if hostbuf is not None else None)

    def __repr__(self):
        return f'<GPUBuffer with shape {self.shape}>'

def compute_output_size(x, y):
    n_dims = max(len(x.shape), len(y.shape))
    x_shape, y_shape = np.ones(n_dims, dtype=np.int32), np.ones(n_dims, dtype=np.int32)
    x_shape[:len(x.shape)] = np.array(x.shape, dtype=np.int32)
    y_shape[:len(y.shape)] = np.array(y.shape, dtype=np.int32)

    if not np.all((x_shape == 1) | (y_shape == 1) | (x_shape == y_shape)):
        raise Exception(f"Unbroadcastable shapes: {x.shape} vs {y.shape}")

    ret_shape = np.maximum(x_shape, y_shape)
    return n_dims, ret_shape, x_shape, y_shape

def compute_broadcasted_dimensions(n_dims, x_shape, y_shape):
    dimension_list, comp_list = [], []

    def push(dim, comp):
        if len(comp_list) > 0 and comp_list[-1] == comp:
            dimension_list[-1] *= dim
        elif comp != (False, False):
            dimension_list.append(dim)
            comp_list.append(comp)
    
    for i in range(n_dims):
        push(np.int32(max(x_shape[i], y_shape[i])), 
             (x_shape[i] > 1, y_shape[i] > 1))
    return dimension_list, comp_list

def unbroadcast(ctx, queue, out, in_shape):
    sum_axis = [i for i in range(len(in_shape)) if in_shape[i]==1 and out.shape[i]>1] if in_shape != (1,) else None
    return reduce_op(ctx, queue, "out += a", "out", out, sum_axis, keepdims=True)

@functools.lru_cache()
def get_binary_op_kernel(ctx, code, complist):
    ndims = len(complist)
    # Arguments for the kernel function. E.g.: int d0, int d1, int d2, int p0, int p1
    # E.g. (cont'd): d0, d1, d2 are the dimensions of the tensor
    # E.g. (cont'd): p0, p1 are the cumulative dimensions in reverse order of the tensor
    args = "".join([f", int d{i}" for i in range(ndims)]) + "".join([f", int p{i}" for i in range(ndims-1)])

    # Computes the indices of the element in the tensor
    # Remember that OpenCL flattened the multi-dimensional tensor into a 1d vector
    compute_idx_rets = ["\n    int idx_ret" + str(i) + " = (gid0 / " +
                        (f"p{i}" if i < ndims - 1 else "1") + ") % d" + 
                        str(i) + ";" for i in range(ndims)]

    # Computes the actual index in the flattened OpenCL tensor
    # Recursive formula: idx_ret2 + d2*(idx_ret1 + d1*(idx_ret0 + d0*(0)))
    idx_exprs = ["0", "0"] # [idx_x, idx_y]
    for i in range(ndims):
        for j in range(2):
            if complist[i][j]:
                idx_exprs[j] = "idx_ret%d + d%d*(%s)" % (i, i, idx_exprs[j])

    return cl.Program(ctx, """__kernel void binop(__global const float *x_g, __global const float *y_g, __global float *res_g"""+args+""") {
    int gid0 = get_global_id(0);"""+"".join(compute_idx_rets)+"""
    float a = x_g["""+idx_exprs[0]+"""];
    float b = y_g["""+idx_exprs[1]+"""];
    res_g[gid0] = """+code+""";\n}""").build()

@functools.lru_cache()
def get_unary_op_kernel(ctx, code):
    return cl.Program(ctx, """__kernel void unop(__global const float *x_g, __global float *res_g) {
        int gid = get_global_id(0);
        float a = x_g[gid];
        res_g[gid] = """+code+""";
    }""").build()

@functools.lru_cache()
def get_reduce_op_kernel(ctx, code, code2, start):
    return cl.Program(ctx, """
    __kernel void reduce(__global const float *a_g, int sz, __global float *res_g, int prod, int n_dims,
                       __global const int *shape_x, __global const int *shape_ret) {
        int gid = get_global_id(0);
        float out = """+start+""";
        for (int x = 0; x < sz; x++) {
        int idx = 0;  // compute index into a_g
        int tprod = prod;
        int tsz = sz;
        for (int dim = 0; dim < n_dims; dim++) {
            idx *= shape_x[dim];
            if (shape_x[dim] == shape_ret[dim]) {   // dim from gid, don't reduce
            tprod /= shape_x[dim];
            idx += (gid / tprod) % shape_x[dim];
            } else {  // dim from x
            tsz /= shape_x[dim];
            idx += (x / tsz) % shape_x[dim];
            }
        }
        float a = a_g[idx];
        """+code+""";
        }
        res_g[gid] = """+code2+""";
  }""").build()

def element_wise_binary_op(ctx, queue, code, x, y):
    n_dims, ret_shape, x_shape, y_shape = compute_output_size(x, y)
    dimension_list, comp_list = compute_broadcasted_dimensions(n_dims, x_shape, y_shape)

    prgm = get_binary_op_kernel(ctx, code, tuple(comp_list))

    out = GPUBuffer(ctx, ret_shape, hostbuf=np.zeros(ret_shape, dtype=np.float32))

    prod_list = np.array(dimension_list, dtype=np.int32)[-1::-1].cumprod(dtype=np.int32)[-1::-1]
    prgm.binop(queue, [prod_list[0]] if len(dimension_list) > 0 else [1], None, 
               x.cl, y.cl, out.cl, *dimension_list, *(prod_list[1:]))
    return out

def unary_op(ctx, queue, code, x):
    out = GPUBuffer(ctx, x.shape, hostbuf=np.zeros(x.shape, dtype=np.float32))
    prgm = get_unary_op_kernel(ctx, code)
    prgm.unop(queue, [np.prod(out.shape)], None, x.cl, out.cl)
    return out

def reduce_op(ctx, queue, code, code2, inp, axis=None, keepdims=False, start="0.0"):
    if isinstance(axis, int): axis = [axis]
    if axis is None:
        osize = [1]*len(inp.shape)
    else:
        osize = np.array(inp.shape)
        osize[list(axis)] = 1

    ret = GPUBuffer(ctx, osize if keepdims else osize[osize != 1], hostbuf=None)

    if axis is None: 
        ret.shape = (1,)
    
    prgm = get_reduce_op_kernel(ctx, code, code2, start)

    prgm.reduce(queue, [np.prod(osize)], None, inp.cl,
        np.int32(np.prod(inp.shape) // np.prod(osize)), ret.cl,
        np.int32(np.prod(osize)), np.int32(len(osize)),
        cl.Buffer(ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR, 
                  hostbuf=np.array(inp.shape, dtype=np.int32)),
        cl.Buffer(ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR, 
                  hostbuf=np.array(osize, dtype=np.int32))
    )
    return ret

def perm_axis_op(ctx, queue, inp, order=(1,0)):
    out_size = np.array(inp.shape)[list(order)]
    ret = GPUBuffer(ctx, out_size)
    prgm = cl.Program(ctx, """
        __kernel void perm(__global const float *a_g, __global float *res_g, int n_axis,
                            __global const int *shape, __global const int *order) {
            int gid = get_global_id(0);
            int gi = gid;
            int idx = 0;
            for(int i = n_axis-1; i>-1; i--) {
            int stride = 1;
            for(int j=order[i]+1; j<n_axis; j++) stride *= shape[j];
            idx += (gi % shape[order[i]])*stride;
            gi /= shape[order[i]];
            }
            res_g[gid] = a_g[idx];
            }"""
    ).build()

    prgm.perm(queue, [np.prod(out_size)], None, inp.cl, ret.cl, np.int32(len(out_size)),
         cl.Buffer(ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR, hostbuf=np.array(inp.shape, dtype=np.int32)),
         cl.Buffer(ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR, hostbuf=np.array(order, dtype=np.int32)))
    return ret

def inner_slice(ctx, queue, inp, arg):
    shift = [y[0] for y in arg]
    out_shape = [y[1]-y[0] for y in arg]
    ret = GPUBuffer(ctx, out_shape)
    prgm = cl.Program(ctx, """
        __kernel void gslice(__global const float *input, __global float *output, int prod, int n_dims,
                       __global const int *shape_x, __global const int *shape_ret,
                       __global const int *shift) {
            int gid = get_global_id(0);
            int iptr = 0;
            int zero = 1;
            for (int dim = 0; dim < n_dims; dim++) {
            prod /= shape_ret[dim];
            int sidx = (gid / prod) % shape_ret[dim] + shift[dim];
            zero &= (sidx >= 0 && sidx < shape_x[dim]);
            iptr = (iptr * shape_x[dim]) + sidx;
            }
            output[gid] = zero ? input[iptr] : 0.0;
        }
    """).build()

    prgm.gslice(queue, [np.prod(ret.shape)], None, inp.cl, ret.cl, np.int32(np.prod(ret.shape)), np.int32(len(ret.shape)),
                cl.Buffer(ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR, hostbuf=np.array(inp.shape, dtype=np.int32)),
                cl.Buffer(ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR, hostbuf=np.array(ret.shape, dtype=np.int32)),
                cl.Buffer(ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR, hostbuf=np.array(shift, dtype=np.int32)))
    return ret

def matmul_op(ctx, queue, a, b):
    cnt = np.prod(a.shape[0:-2]) if len(a.shape) > 2 else 1
    isize, msize, osize = np.int32(a.shape[-2]), np.int32(a.shape[-1]), np.int32(b.shape[-1])
    ret = GPUBuffer(ctx, list(a.shape[0:-2])+[isize, osize])
    prgm = cl.Program(ctx, """
        __kernel void matmul(__global const float *input, __global const float *weight, __global float *res,
                            int isize, int is0, int is1, int msize, int ws0, int ws1, int osize) {
            int stride = get_global_id(2);
            int X = get_global_id(0); // isize
            int Y = get_global_id(1); // osize
            float ret = 0.0;
            for (int x = 0; x < msize; x++) {
                ret += input[X * is0 + x * is1 + isize*msize*stride] *
                weight[Y * ws0 + x * ws1 + msize*osize*stride];
            }
            res[X * osize + Y + isize*osize*stride] = ret;
        }
    """).build()

    prgm.matmul(queue, [isize, osize, cnt], None, 
                a.cl, b.cl, ret.cl, isize, msize, 
                np.int32(1), msize, np.int32(1), osize, osize)
    return ret

def one_hot_encoding(ctx, queue, a, num_classes):
    ret = GPUBuffer(ctx, (a.shape[0], num_classes))
    n_rows, n_cols = np.int32(a.shape[0]), np.int32(num_classes)
    prgm = cl.Program(ctx, """
        __kernel void one_hot(__global const float *labels, __global float *res, 
                              const int n_cols) {
            int gid0 = get_global_id(0);

            int label = labels[gid0];
            res[gid0 * n_cols + label] = 1;
        }
    """).build()

    prgm.one_hot(queue, [n_rows, n_cols], None, a.cl, ret.cl, n_cols)
    return ret

def pad1d_op(ctx, queue, a, pad, length):
    padded_length = np.int32(length + 2 * pad)
    batch_size, in_channel = np.int32(a.shape[0]), np.int32(a.shape[1])
    ret = GPUBuffer(ctx, (batch_size, in_channel, padded_length))
    prgm = cl.Program(ctx, """
        __kernel void pad1d(__global const float *a, __global float *res, const int batch_size,
                            const int in_channel, const int padded_length, const int length,
                            const int pad) {
            
            int batch_id = get_global_id(0);
            int channel_id = get_global_id(1);
            int col = get_global_id(2);

            if((col - pad >= 0) && (col - pad < length)) {
                res[batch_id * in_channel * padded_length + channel_id * padded_length + col] = \
                  a[batch_id * in_channel * length + channel_id * length + col-pad];
            }
        }
    """).build()

    prgm.pad1d(queue, [batch_size, in_channel, padded_length], None, a.cl, ret.cl,
               batch_size, in_channel, padded_length, np.int32(length), np.int32(pad))
    return ret

def pad2d_op(ctx, queue, a, pad, im_height, im_width):
    padded_height, padded_width = np.int32(im_height + 2 * pad), np.int32(im_width + 2 * pad)
    batch_size, in_channel = np.int32(a.shape[0]), np.int32(a.shape[1])
    ret = GPUBuffer(ctx, (batch_size, in_channel, padded_height, padded_width))
    prgm = cl.Program(ctx, """
        __kernel void pad2d(__global const float *a, __global float *res, const int batch_size, 
                            const int in_channel, const int padded_height, const int padded_width,
                            const int im_height, const int im_width, const int pad) {

            int batch_id = get_global_id(0) / in_channel;
            int channel_id = get_global_id(0) % in_channel;
            int row = get_global_id(1);
            int col = get_global_id(2);

            if((row - pad >= 0) && (col - pad >= 0) && (col - pad < im_width) && (row - pad < im_height)) {
                res[batch_id * in_channel * padded_height * padded_width + channel_id * padded_height * padded_width + row * padded_width + col] = \
                  a[batch_id * in_channel * im_height * im_width + channel_id * im_height * im_width + (row - pad) * im_width + col-pad];
            }
        }
    """).build()
    
    prgm.pad2d(queue, [batch_size * in_channel, padded_height, padded_width], 
               None, a.cl, ret.cl, batch_size, in_channel, padded_height, padded_width,
               np.int32(im_height), np.int32(im_width), np.int32(pad))
    return ret

def conv1d_op(ctx, queue, a, weight, stride, output_length):
    batch_size, stride = np.int32(a.shape[0]), np.int32(stride)
    output_length = np.int32(output_length)
    num_filters, in_channel = np.int32(weight.shape[0]), np.int32(weight.shape[1])
    kernel_length = np.int32(weight.shape[2])
    length = np.int32(a.shape[2])

    ret = GPUBuffer(ctx, (batch_size, num_filters, output_length))
    prgm = cl.Program(ctx, """
        __kernel void conv1d(__global const float *input, __global const float *weight,
                             __global float *output, const int kernel_length, const int num_filters, const int in_channel,
                             const int out_length, const int length, const int stride) {
            int batch_id = get_global_id(0);
            int filter_id = get_global_id(1);
            int col = get_global_id(2);

            float res = 0.0;

            for(int c = 0; c < in_channel; c++) {
                for (int x = col * stride; x < col * stride + kernel_length; x++) {
                    res += input[batch_id * in_channel * length + c * length + x] * \
                           weight[filter_id * in_channel * kernel_length + c * kernel_length + \
                                  (x - col * stride)];
                }
            }

            output[batch_id * num_filters * out_length + filter_id * out_length + col] = res;
        }
    """).build()

    prgm.conv1d(queue, [batch_size, num_filters, output_length], None, a.cl, weight.cl, 
                ret.cl, kernel_length, num_filters, in_channel, np.int32(output_length), length, 
                stride)
    return ret

def conv2d_op(ctx, queue, a, weight, stride, output_height, output_width):
    batch_size, stride = np.int32(a.shape[0]), np.int32(stride)
    output_height, output_width = np.int32(output_height), np.int32(output_width)
    num_filters, in_channel = np.int32(weight.shape[0]), np.int32(weight.shape[1])
    kernel_height, kernel_width  = np.int32(weight.shape[2]), np.int32(weight.shape[3])
    im_height, im_width = np.int32(a.shape[2]), np.int32(a.shape[3])

    ret = GPUBuffer(ctx, (batch_size, num_filters, output_height, output_width))
    prgm = cl.Program(ctx, """
        __kernel void conv2d(__global const float *input, __global const float *weight,
                             __global float *output, const int kernel_height, const int kernel_width, const int num_filters, 
                             const int in_channel, const int out_height, const int out_width, const int im_height, 
                             const int im_width, const int stride) {
            int gid0 = get_global_id(0);
            int row_id = get_global_id(1);
            int col_id = get_global_id(2);

            int batch_id = gid0 / num_filters;
            int filter_id = gid0 % num_filters;

            float res = 0.0;

            for(int c = 0; c < in_channel; c++) {
                for(int y = row_id * stride; y < row_id * stride + kernel_height; y++) {
                    for(int x = col_id * stride; x < col_id * stride + kernel_width; x++) {
                        res += input[batch_id*in_channel*im_height*im_width + c*im_height*im_width + y*im_width + x] * \
                               weight[filter_id*in_channel*kernel_height*kernel_width + c*kernel_height*kernel_width + \
                                      (y-row_id*stride)*kernel_width + (x-col_id*stride)];
                    }
                }
            }

           output[batch_id * num_filters * out_height * out_width + filter_id * out_height * out_width + row_id * out_width + col_id] = res; 
        }
    """).build()

    prgm.conv2d(queue, [batch_size * num_filters, output_height, output_width], None, a.cl, weight.cl,
                ret.cl, kernel_height, kernel_width, num_filters, in_channel, output_height, output_width, im_height,
                im_width, stride)
    return ret


# *************************************
# *********** Forward passes **********
# *************************************

def squeeze_forward(ctx, queue, a, axis):
    in_shape = np.array(a.shape)
    out_shape = np.delete(in_shape, axis).astype(np.int32)
    return GPUBuffer(ctx, out_shape, hostbuf=a.data)

def unsqueeze_forward(ctx, queue, a, axis):
    in_shape = np.array(a.shape)
    out_shape = np.insert(in_shape, axis, 1).astype(np.int32)
    return GPUBuffer(ctx, out_shape, hostbuf=a.data)
    
def add_forward(ctx, queue, a, b):
    return element_wise_binary_op(ctx, queue, 'a+b', a, b)

def mul_forward(ctx, queue, a, b):
    return element_wise_binary_op(ctx, queue, 'a*b', a, b)

def matmul_forward(ctx, queue, a, b):
    return matmul_op(ctx, queue, a, b)

def log_forward(ctx, queue, a):
    out = unary_op(ctx, queue, 'log(a)', a)
    return out

def exp_forward(ctx, queue, a):
    return unary_op(ctx, queue, 'exp(a)', a)

def neg_forward(ctx, queue, a):
    return unary_op(ctx, queue, '-a', a)

def pow_forward(ctx, queue, a, exp):
    return unary_op(ctx, queue, f'pow(a, (float){exp})', a)

def relu_forward(ctx, queue, a):
    return unary_op(ctx, queue, 'max(a, (float)0.)', a)

def sigmoid_forward(ctx, queue, a):
    return unary_op(ctx, queue, '1.0 / (1 + exp(-a))', a)

def tanh_forward(ctx, queue, a):
    return unary_op(ctx, queue, '(exp(a) - exp(-a)) / (exp(a) + exp(-a))', a)

def slice_forward(ctx, queue, a, indices):
    return inner_slice(ctx, queue, a, indices)

def transpose_forward(ctx, queue, a):
    return perm_axis_op(ctx, queue, a, order=(1, 0))

def reshape_forward(ctx, queue, a, shape):
    shape = tuple(-np.prod(a.shape) // np.prod(shape) if s == -1 else s for s in shape)
    return GPUBuffer(ctx, shape, hostbuf=a.data)

def max_forward(ctx, queue, a, axis):
    return reduce_op(ctx, queue, 'out = max(a, out)', 'out',
                     a, axis=axis, keepdims=False, start='-INFINITY')

def min_forward(ctx, queue, a, axis):
    return reduce_op(ctx, queue, 'out = min(a, out)', 'out',
                     a, axis=axis, keepdims=False, start='+INFINITY')

def sum_forward(ctx, queue, a, axis):
    return reduce_op(ctx, queue, 'out += a', 'out', 
                     a, axis=axis, keepdims=False)

def conv1d_forward(ctx, queue, a, weight, stride):
    batch_size, in_channel, length = a.shape
    num_filters, _, kernel_length = weight.shape
    output_length = get_conv1d_output_size(length, kernel_length, stride, 0)

    out = conv1d_op(ctx, queue, a, weight, stride, output_length)
    return out

def conv2d_forward(ctx, queue, a, weight, stride):
    batch_size, in_channel, im_height, im_width = a.shape
    num_filters, _, kernel_height, kernel_width = weight.shape

    output_height, output_width = get_conv2d_output_size(im_height, im_width,
        (kernel_height, kernel_width), stride, 0)

    out = conv2d_op(ctx, queue, a, weight, stride, output_height, output_width)
    return out

# *************************************
# ********** Backward passes **********
# *************************************

def squeeze_backward(ctx, queue, grad_output, axis):
    in_shape = np.array(grad_output.shape)
    out_shape = np.insert(in_shape, axis, 1).astype(np.int32)
    return GPUBuffer(ctx, out_shape, hostbuf=grad_output.data)

def unsqueeze_backward(ctx, queue, grad_output, axis):
    in_shape = np.array(grad_output.shape)
    out_shape = np.delete(in_shape, axis).astype(np.int32)
    return GPUBuffer(ctx, out_shape, hostbuf=grad_output.data)

def add_backward(ctx, queue, grad_output, a_shape, b_shape):
    return unbroadcast(ctx, queue, grad_output, a_shape), unbroadcast(ctx, queue, grad_output, b_shape) 

def mul_backward(ctx, queue, grad_output, a, b):
    grad_a = element_wise_binary_op(ctx, queue, 'a*b', grad_output, b)
    grad_b = element_wise_binary_op(ctx, queue, 'a*b', grad_output, a)
    return unbroadcast(ctx, queue, grad_a, a.shape), unbroadcast(ctx, queue, grad_b, b.shape)

def matmul_backward(ctx, queue, grad_output, a, b):
    aT = perm_axis_op(ctx, queue, a)
    bT = perm_axis_op(ctx, queue, b)
    grad_a = matmul_op(ctx, queue, grad_output, bT)
    grad_b = matmul_op(ctx, queue, aT, grad_output)
    return grad_a, grad_b

def log_backward(ctx, queue, grad_output, a):
    return element_wise_binary_op(ctx, queue, 'a/b', grad_output, a)

def exp_backward(ctx, queue, grad_output, a):
    return element_wise_binary_op(ctx, queue, 'a * exp(b)', grad_output, a)

def neg_backward(ctx, queue, grad_output):
    return unary_op(ctx, queue, '-a', grad_output)

def pow_backward(ctx, queue, grad_output, a, exp):
    return element_wise_binary_op(ctx, queue, f'a * {exp} * pow(b, (float){exp-1})', grad_output, a)

def relu_backward(ctx, queue, grad_output, a):
    return element_wise_binary_op(ctx, queue, 'a*(b>=0)', grad_output, a)

def sigmoid_backward(ctx, queue, grad_output, a):
    return element_wise_binary_op(ctx, queue, 'a*(exp(-b) / pow((1 + exp(-b)), 2))', 
                                  grad_output, a)

def tanh_backward(ctx, queue, grad_output, a):
    return element_wise_binary_op(ctx, queue, 'a*(1 - pow(exp(b) - exp(-b), 2) / pow(exp(b) + exp(-b), 2))', 
                                  grad_output, a)

def slice_backward(ctx, queue, grad_output, shape, fwd_indices):
    indices = [(0 - p[0], grad_output.shape[i] + (shape[i] - p[1])) for i, p in enumerate(fwd_indices)]
    return inner_slice(ctx, queue, grad_output, indices)

def transpose_backward(ctx, queue, grad_output):
    return perm_axis_op(ctx, queue, grad_output)

def reshape_backward(ctx, queue, grad_output, shape):
    new_shape = tuple([-np.prod(grad_output.shape) // np.prod(shape) 
                       if s == -1 else s for s in shape])
    assert np.prod(new_shape) == np.prod(shape), "Inconsistent array reshape size"
    return GPUBuffer(ctx, new_shape, hostbuf=grad_output)

def max_backward(ctx, queue, grad_output, inp, out, axis):
    shape = [1 if axis is None or i in axis else inp.shape[i] for i in range(len(inp.shape))]
    ret2 = element_wise_binary_op(ctx, queue, "1.0*(a==b)", inp, GPUBuffer(ctx, shape, out))
    div = reduce_op(ctx, queue, "out += a", "out+1e-10", ret2, axis=axis)
    ret3 = element_wise_binary_op(ctx, queue, "a/b", ret2, GPUBuffer(ctx, shape, div))
    return element_wise_binary_op(ctx, queue, 'a*b', ret3, GPUBuffer(ctx, shape, grad_output))

def min_backward(ctx, queue, grad_output, inp, out, axis):
    shape = [1 if axis is None or i in axis else inp.shape[i] for i in range(len(inp.shape))]
    ret2 = element_wise_binary_op(ctx, queue, "1.0*(a==b)", inp, GPUBuffer(ctx, shape, out))
    div = reduce_op(ctx, queue, "out += a", "out+1e-10", ret2, axis=axis)
    ret3 = element_wise_binary_op(ctx, queue, "a/b", ret2, GPUBuffer(ctx, shape, div))
    return element_wise_binary_op(ctx, queue, 'a*b', ret3, GPUBuffer(ctx, shape, grad_output))

def sum_backward(ctx, queue, grad_output, a, axis):
    axis = [axis] if type(axis) == int else axis
    shape = [1 if axis is None or i in axis else a.shape[i] for i in range(len(a.shape))]
    output = GPUBuffer(ctx, shape, hostbuf=grad_output)
    return element_wise_binary_op(ctx, queue, 'a+b', output, GPUBuffer(ctx, a.shape))

def conv1d_backward(ctx, queue, grad_output, x, weight, stride):
    batch_size, _, out_length = grad_output.shape
    num_filters, _, kernel_length = weight.shape
    _, in_channel, in_length = x.shape

    grad_x = GPUBuffer(ctx, shape=x.shape)
    grad_weight = GPUBuffer(ctx, shape=weight.shape)

    prgm_grad_x = cl.Program(ctx, """
        __kernel void conv_backward_x(__global const float *weight, __global const float *grad_output, __global float *grad_x, const int kernel_length,
                                      const int num_filters, const int in_channel, const int out_length, const int in_length, const int stride,
                                      const int batch_size) {

            int batch = get_global_id(0);
            int channel = get_global_id(1);

            for(int x = 0; x < out_length; x++) {
                for(int kx = 0; kx < kernel_length; kx++) {

                    float sum = 0.0;

                    for(int f = 0; f < num_filters; f++) {
                        sum += grad_output[batch * num_filters * out_length + f * out_length + x] * \
                            weight[f * in_channel * kernel_length + channel * kernel_length + kx];
                    } 

                    grad_x[batch * in_channel * in_length + channel * in_length + x * stride + kx] += sum;
                }
            }


        }
    """).build()

    prgm_grad_weight = cl.Program(ctx, """
        __kernel void conv_backward_weight(__global const float *x_tensor, __global const float *grad_output, __global float *grad_weight,
                                           const int kernel_length, const int num_filters, const int in_channel, const int out_length, 
                                           const int in_length, const int stride, const int batch_size) {

            int filter = (get_global_id(0) / in_channel) % num_filters;
            int channel = get_global_id(0) % in_channel;
            int col = get_global_id(1);

            float sum = 0.0;

            for(int x = 0; x < out_length; x++) {
                for(int b = 0; b < batch_size; b++) {
                    sum += grad_output[b * num_filters * out_length + filter * out_length + x] * \
                        x_tensor[b * in_channel * in_length + channel * in_length + x * stride + col];
                }
            }
            grad_weight[get_global_id(0) * kernel_length + col] = sum;

        }
    """).build()

    grad_x_args = (weight.cl, grad_output.cl, grad_x.cl, np.int32(kernel_length), np.int32(num_filters), 
                   np.int32(in_channel), np.int32(out_length), np.int32(in_length), np.int32(stride), 
                   np.int32(batch_size))
    
    grad_w_args = (x.cl, grad_output.cl, grad_weight.cl, np.int32(kernel_length), np.int32(num_filters), 
                   np.int32(in_channel), np.int32(out_length), np.int32(in_length), np.int32(stride), 
                   np.int32(batch_size))

    prgm_grad_x.conv_backward_x(queue, [batch_size, in_channel], None, *grad_x_args)
    prgm_grad_weight.conv_backward_weight(queue, [num_filters * in_channel, kernel_length], None, *grad_w_args)

    return grad_x, grad_weight
    
    

def conv2d_backward(ctx, queue, grad_output, x, weight, stride):
    batch_size, _, out_height, out_width = grad_output.shape
    num_filters, _, kernel_height, kernel_width = weight.shape
    _, in_channel, im_height, im_width = x.shape

    grad_x = GPUBuffer(ctx, shape=x.shape)
    grad_weight = GPUBuffer(ctx, shape=weight.shape)

    prgm_grad_x = cl.Program(ctx, """
        __kernel void conv_backward_x(__global const float *weight, __global const float *grad_output, __global float *grad_x, const int kernel_height,
                                      const int kernel_width, const int num_filters, const int in_channel, const int out_width, const int out_height,
                                      const int im_width, const int im_height, const int stride, const int batch_size) {

            int batch = get_global_id(0);
            int channel = get_global_id(1);

            for(int y = 0; y < out_height; y++) {
                for(int x = 0; x < out_width; x++) {
                    for(int ky = 0; ky < kernel_height; ky++) {
                        for(int kx = 0; kx < kernel_width; kx++) {
                            
                            float sum = 0.0;

                            for(int f = 0; f < num_filters; f++) {
                                sum += grad_output[batch * num_filters * out_height * out_width + f * out_height * out_width + y * out_width + x] * \
                                    weight[f * in_channel * kernel_height * kernel_width + channel * kernel_height * kernel_width + ky * kernel_width + kx];
                            }

                            grad_x[batch * in_channel * im_height * im_width + channel * im_height * im_width + (y * stride + ky) * im_width + x * stride + kx] += sum;
                        }
                    }
                }
            }
        }
    """).build()

    prgm_grad_weight = cl.Program(ctx, """
        __kernel void conv_backward_weight(__global const float *x_tensor, __global const float *grad_output, __global float *grad_weight,
                                           const int kernel_height, const int kernel_width, const int num_filters, const int in_channel,
                                           const int out_height, const int out_width, const int im_height, const int im_width, const int stride,
                                           const int batch_size) {

            int filter = (get_global_id(0) / in_channel) % num_filters;
            int channel = get_global_id(0) % in_channel;
            int row = get_global_id(1);
            int col = get_global_id(2);

            float sum = 0.0;

            for(int y = 0; y < out_height; y++) {
                for(int x = 0; x < out_width; x++) {
                    for(int b = 0; b < batch_size; b++) {
                        sum += grad_output[b * num_filters * out_height * out_width + filter * out_height * out_width + y * out_width + x] * \
                            x_tensor[b * in_channel * im_width * im_height + channel * im_height * im_width + im_width * (y * stride + row) + x * stride + col];
                    }
                }
            }
            grad_weight[get_global_id(0)*kernel_height*kernel_width + row * kernel_width + col] = sum;
        }
    """).build()

    grad_x_args = (weight.cl, grad_output.cl, grad_x.cl, np.int32(kernel_height), np.int32(kernel_width), np.int32(num_filters), np.int32(in_channel),
                   np.int32(out_width), np.int32(out_height), np.int32(im_width), np.int32(im_height), np.int32(stride), np.int32(batch_size))

    grad_w_args = (x.cl, grad_output.cl, grad_weight.cl, np.int32(kernel_height), np.int32(kernel_width), np.int32(num_filters), np.int32(in_channel),
                   np.int32(out_height), np.int32(out_width), np.int32(im_height), np.int32(im_width), np.int32(stride), np.int32(batch_size))

    prgm_grad_x.conv_backward_x(queue, [batch_size, in_channel], None, *grad_x_args)
    prgm_grad_weight.conv_backward_weight(queue, [num_filters * in_channel, kernel_height, kernel_width], None, *grad_w_args)

    return grad_x, grad_weight


import numpy as np


def get_conv1d_output_size(input_length:int, 
                           kernel_size:int, 
                           stride:int, 
                           padding:int) -> int:
    """Computes the output size of a 1d convolution.

    Args:
        input_length (int): Length of the sequence
        kernel_size (int): Spatial size of the kernel
        stride (int): Stride of the convolution
        padding (int): Spacing between kernel elements

    Returns:
        int: Output size
    """
    input_length += 2 * padding
    return int((input_length - kernel_size) // stride + 1)

def get_conv2d_output_size(input_height:int, input_width:int, 
                           kernel_size:int, stride:int, 
                           padding:int) -> tuple:
    """Computs the output size of a 2d convolution.

    Args:
        input_height (int): Height of the input tensor
        input_width (int): Width of the input tensor
        kernel_size (int): Square size of the kernel
        stride (int): Stride of te convolution
        padding (int): zero-padding added to both sides of the input

    Returns:
        tuple: output height and output width
    """
    input_height += 2 * padding
    input_width += 2 * padding

    output_height = (input_height - kernel_size[0]) // stride + 1
    output_width = (input_width - kernel_size[1]) // stride + 1
    return int(output_height), int(output_width)

def get_im2col_indices(x_shape:tuple, field_height:int, 
                       field_width:int, pad:int, stride:int) -> tuple:
    """Performs a im2col transformation for fast backward passes in 2d
       convolution layer

    Args:
        x_shape (int): Shape of the input tensor
        field_height (int): Height of the kernel
        field_width (int): Width of the kernel
        pad (int): Amount of zero-padding applied in the convolution
        stride (int): Stride of the convolution

    Returns:
        tuple: indices for im2col
    """
    N, C, H, W = x_shape
    assert (H + 2 * pad - field_height) % stride == 0
    assert (W + 2 * pad - field_height) % stride == 0
    out_height, out_width = get_conv2d_output_size(H, W, (field_height, field_width), stride, pad)

    i0 = np.repeat(np.arange(field_height), field_width)
    i0 = np.tile(i0, C)
    i1 = stride * np.repeat(np.arange(out_height), out_width)
    j0 = np.tile(np.arange(field_width), field_height * C)
    j1 = stride * np.tile(np.arange(out_width), out_height)
    i = i0.reshape(-1, 1) + i1.reshape(1, -1)
    j = j0.reshape(-1, 1) + j1.reshape(1, -1)

    k = np.repeat(np.arange(C), field_height * field_width).reshape(-1, 1)

    return (k, i, j)

def col2im(cols:np.ndarray, x_shape:tuple, field_height:int, 
           field_width:int, padding:int, stride:int) -> np.ndarray:
    """Performs a col2im transformation for fast backward passes in 2d


    Args:
        cols (np.ndarray): Input array transformed using im2col
        x_shape (tuple): Shape of the input
        field_height (int): Height of the kernel
        field_width (int): Width of the kernel
        padding (int): Amount of zero-padding applied in the convolution
        stride (int): Stride of the convolution

    Returns:
        np.ndarray: image transformed
    """

    N, C, H, W = x_shape
    H_padded, W_padded = H + 2 * padding, W + 2 * padding
    x_padded = np.zeros((N, C, H_padded, W_padded), dtype=cols.dtype)
    k, i, j = get_im2col_indices(x_shape, field_height, field_width, padding,
                                 stride)
    cols_reshaped = cols.reshape(C * field_height * field_width, -1, N)
    cols_reshaped = cols_reshaped.transpose(2, 0, 1)
    np.add.at(x_padded, (slice(None), k, i, j), cols_reshaped)
    if padding == 0:
        return x_padded
    return x_padded[:, :, padding:-padding, padding:-padding]


from nanograd.tensor import Tensor
import nanograd.tensor as tensor
from nanograd.device import Device

import numpy as np

from typing import Union


def init_weights(shape:tuple, weight_initialization:str, 
                 fan_mode:str="fan_in", **kwargs) -> Tensor:
    """Initializes the weights of a layer

    Args:
        shape (tuple): Shape of the tensor
        weight_initialization (str): Method of initialization (Glorot or Kaiming)
        fan_mode (str, optional): Fan mode. Defaults to "fan_in".

    Raises:
        Exception: Method must be Kaiming or Xavier normal or uniform 

    Returns:
        Tensor: Tensor containing the initialized weights
    """
    assert type(shape) == tuple, f"Shape must be a tuple. Got {type(shape).__name__}."
    assert fan_mode in ["fan_in", "fan_out"], "Wrong fan mode. Only fan_in and fan_out supported."

    out_features, in_features = shape[0], shape[1]

    if weight_initialization == "kaiming_normal":
        std = np.sqrt(1.0 / in_features) if fan_mode == "fan_in" else np.sqrt(1.0 / out_features)
        return Tensor.normal(0.0, std, shape, **kwargs)

    elif weight_initialization == "kaiming_uniform":
        bound = np.sqrt(3.0 / in_features) if fan_mode == "fan_in" else np.sqrt(3.0 / out_features)
        weight = np.random.uniform(-bound, bound, shape)
        return Tensor(weight, **kwargs)

    elif weight_initialization == "glorot_normal":
        std = np.sqrt(2.0 / (in_features + out_features))
        return Tensor.normal(0.0, std, shape, **kwargs)

    elif weight_initialization == "glorot_uniform":
        bound = np.sqrt(6.0 / (in_features + out_features))
        weight = np.random.uniform(-bound, bound, shape) 
        return Tensor(weight, **kwargs)

    else:
        raise Exception("Unknown weight initialization methods. Only Glorot and Kaiming are available.")


class Module:
    """Base class (superclass) for all components of an NN.

    Layer classes and even full Model classes should inherit from this Module.
    Inheritance gives the subclass all the functions/variables below

    """

    def __init__(self) -> None:
        self._submodules = {} # Submodules of the class
        self._parameters = {} # Trainable parameters in modules and its submodules
        self._tensors = {}

        self.is_train = True
    
    def train(self) -> None:
        """Activates training mode for network component"""
        self.is_train = True
        return self

    def eval(self) -> None:
        """Activates evaluation mode for network component"""
        self.is_train = False
        return self
    
    def forward(self, *args):
        """Forward pass of the module"""
        raise NotImplementedError("Subclasses of module should have a forward method implemented")

    def is_parameter(self, obj) -> bool:
        """Checks if input object is a Tensor of trainable params"""
        return isinstance(obj, Tensor) and obj.is_parameter

    def parameters(self):
        """
            Returns an interator over stored params.
            Includes submodules' params too
        """
        self._ensure_is_initialized()
        for name, parameter in self._parameters.items():
            yield parameter
        for name, module in self._submodules.items():
            for parameter in module.parameters():
                yield parameter
    
    def add_parameter(self, name, value) -> None:
        """Stores params"""
        self._ensure_is_initialized()
        self._parameters[name] = value

    def add_module(self, name, value) -> None:
        """Stores module and its params"""
        self._ensure_is_initialized()
        self._submodules[name] = value
    
    def add_tensor(self, name, value) -> None:
        """Stores tensors"""
        self._ensure_is_initialized()
        self._tensors[name] = value
    
    def __setattr__(self, name, value):
        """Stores params or modules that you provide"""
        if self.is_parameter(value):
            self.add_parameter(name, value)
            self.add_tensor(name, value)
        elif isinstance(value, Module):
            self.add_module(name, value)
        elif isinstance(value, Tensor):
            self.add_tensor(name, value)

        object.__setattr__(self, name, value)

    def __call__(self, *args):
        """Runs self.forward(args)"""
        return self.forward(*args)

    def _ensure_is_initialized(self):
        """Ensures that subclass's __init__() method ran super().__init__()"""
        if self.__dict__.get('_submodules') is None:
            raise Exception("Modules not initialized. Did not initialize the parent class. \
                             Call super().__init__().")
        if self.__dict__.get('_tensors') is None:
            raise Exception("Tensors not initialized. Did not initialize the parent class. \
                             Call super().__init__().")
        if self.__dict__.get('_parameters') is None: 
            raise Exception("Parameters not initialized. Did not initialize the parent class. \
                             Call super().__init__().")
    
    def cpu(self):
        """Moving all tensors onto the CPU"""
        if len(self._tensors) > 0:
            for tensor in self._tensors.items():
                tensor[1].cpu()
        if len(self._submodules) > 0:
            for module in self._submodules.items():
                module[1].cpu()
        return self
        
    def gpu(self):
        """Moving all tensors onto the GPU"""
        if len(self._tensors) > 0:
            for tensor in self._tensors.items():
                tensor[1].gpu()
        if len(self._submodules) > 0:
            for module in self._submodules.items():
                module[1].gpu()
        return self


class Sequential(Module):
    """Passes input data through stored layers, in order

    >>> model = Sequential(Linear(2,3), ReLU())
    >>> model(x)
    <output after linear then relu>

    Inherits from:
        Module (nn.module.Module)
    """
    def __init__(self, *layers) -> None:
        super().__init__()
        self.layers = layers

        for idx, l in enumerate(self.layers):
            self.add_module(str(idx), l)

    def __iter__(self):
        """Enables list-like iteration through layers"""
        yield from self.layers

    def __getitem__(self, idx:int):
        """Enables list-like indexing for layers"""
        return self.layers[idx]

    def train(self) -> None:
        """Sets this object and all trainable modules within to train mode"""
        self.is_train = True
        for submodule in self._submodules.values():
            submodule.train()

    def eval(self) -> None:
        """Sets this object and all trainable modules within to eval mode"""
        self.is_train = False
        for submodule in self._submodules.values():
            submodule.eval()

    def forward(self, x:Tensor) -> Tensor:
        """Passes input data through each layer in order
            
            Args:
                x (Tensor): Input data
            Returns:
                Tensor: Output after passing through layers
        """
        for layer in self.layers: 
            x = layer(x)
        return x
    
    def gpu(self):
        for layer in self.layers:
            layer.gpu()
        return self
    
    def cpu(self):
        for layer in self.layers:
            layer.cpu()
        return self

class Linear(Module):
    """A linear layer (aka 'fully-connected' or 'dense' layer)

        >>> layer = Linear(2,3)
        >>> layer(Tensor.ones(10,2)) # (batch_size, in_features)
        <some tensor output with size (batch_size, out_features)>

        Args:
            in_features (int): # dims in input
                            (i.e. # of inputs to each neuron)
            out_features (int): # dims of output
                            (i.e. # of neurons)
            weight_initialization (str): Weight initialization mode
            fan_mode (str): Fan mode as used in the weight initializer
            with_bias (bool): A Linear layer with or without bias

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, in_features:int, out_features:int, 
                 weight_initialization:str="kaiming_normal", fan_mode:str="fan_in",
                 with_bias=True) -> None:
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.with_bias = with_bias

        self.weight = init_weights(
            (self.out_features, self.in_features), weight_initialization, fan_mode,
            requires_grad=True, is_parameter=True, name="lin_weight")
        
        if self.with_bias:
            self.bias = Tensor.zeros((self.out_features, ), requires_grad=True, 
                                    is_parameter=True, name="lin_bias")

    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, in_features)
            Returns:
                Tensor: (batch_size, out_features)
        """
        if self.with_bias:
            out = x @ self.weight.T() + self.bias.unsqueeze(1).T()
        else:
            out = x @ self.weight.T()
           
        out.name = "lin_res"
        return out   


class BatchNorm1d(Module):
    """Batch Normalization Layer 1d

        Args:
            num_features (int): # dims in input and output
            eps (float): Value added to denominator for numerical stability
            momentum (float): Value used for running mean and var computation (smoothing parameter)

        Inherits from:
            Module (mytorch.nn.module.Module)
    """
    def __init__(self, num_features:int, eps:float=1e-5, momentum:float=0.1) -> None:
        super().__init__()
        self.num_features = num_features

        self.eps = eps
        self.momentum = momentum

        self.weight = Tensor.ones((self.num_features,), requires_grad=True, is_parameter=True, name="bn_gamma")
        self.bias = Tensor.zeros((self.num_features,), requires_grad=True, is_parameter=True, name="bn_beta")
        
        self.running_mean = Tensor.zeros((self.num_features,), requires_grad=False, is_parameter=False, name="bn_running_mean")
        self.running_var = Tensor.ones((self.num_features,), requires_grad=False, is_parameter=False, name="bn_running_var")

    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, num_features)
            Returns:
                Tensor: (batch_size, num_features)
        """
        if self.is_train == True:
            batch_mean = x.mean(0)
            batch_var = ((x - batch_mean.reshape(shape=[1, -1])) ** 2).mean(0)
            batch_empirical_var = ((x - batch_mean.reshape(shape=[1, -1])) ** 2).sum(0) / (x.shape[0] - 1)

            self.running_mean = (1. - self.momentum) * self.running_mean + self.momentum * batch_mean
            self.running_var = (1. - self.momentum) * self.running_var + self.momentum * batch_empirical_var

            return self._normalize(x, batch_mean, batch_var)
        else:
            return self._normalize(x, self.running_mean, self.running_var)
        
    def _normalize(self, x:Tensor, mean:Tensor, var:Tensor) -> Tensor:
        """Performs the actual normalization operation

        Args:
            x (Tensor): Input
            mean (Tensor): Computed batch mean if training, else running mean for validation
            var (Tensor): Computed batch var if training, else running var for validation

        Returns:
            Tensor: Normalized tensor
        """
        x_hat = (x - mean.reshape(shape=[1, -1])) / (var + self.eps).sqrt().reshape(shape=[1, -1])
        out = self.weight.reshape(shape=[1, -1]) * x_hat + self.bias.reshape(shape=[1, -1])
        out.name = "bn_1d_res"
        return out


class BatchNorm2d(Module):
    """Batch Normalization Layer 2d

        Args:
            size (tuple): # dims in input and output
            eps (float): value added to denominator for numerical stability
            momentum (float): value used for running mean and var computation (smoothing parameter)

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, num_features:int, eps:float=1e-5, momentum:float=0.1) -> None:
        super().__init__()
        self.num_features = num_features

        self.eps = eps
        self.momentum = momentum

        self.weight = Tensor.ones(self.num_features, requires_grad=True, is_parameter=True, name="bn_gamma")
        self.bias = Tensor.zeros(self.num_features, requires_grad=True, is_parameter=True, name="bn_beta")

        self.running_mean = Tensor.zeros(self.num_features, requires_grad=False, is_parameter=False, name="bn_running_mean")
        self.running_var = Tensor.ones(self.num_features, requires_grad=False, is_parameter=False, name="bn_running_var")

    def forward(self, x:Tensor) -> Tensor:
        """Forward pass

            Args:
                x (Tensor): (batch_size, num_features)
            Returns:
                Tensor: (batch_size, num_features)
        """
        if self.is_train == True: 
            batch_mean = x.mean(axis=(0, 2, 3)) 
            batch_var = ((x - batch_mean.reshape(shape=[1, -1, 1, 1])) ** 2).mean(axis=(0, 2, 3)) 
            batch_empirical_var = (((x - batch_mean.reshape(shape=[1, -1, 1, 1])) ** 2).sum(axis=(0, 2, 3)) / 
                                  (np.prod([x.shape[i] for i in [0, 2, 3]])-1))

            self.running_mean = (1. - self.momentum) * self.running_mean + self.momentum * batch_mean
            self.running_var = (1. - self.momentum) * self.running_var + self.momentum * batch_empirical_var

            return self._normalize(x, batch_mean, batch_var)
        else:
            return self._normalize(x, self.running_mean, self.running_var)
        
    def _normalize(self, x:Tensor, mean:Tensor, var:Tensor) -> Tensor:
        """Normalize a Tensor with mean and variance

            Args:
                x (Tensor): Input
                mean (Tensor): Computed batch mean if training, else running mean for validation
                var (Tensor): Computed batch var if training, else running var for validation
            
            Returns:
                Tensor: Normalized tensor
        """
        x_hat = (x - mean.reshape(shape=[1, -1, 1, 1])) * self.weight.reshape(shape=[1, -1, 1, 1])
        out = x_hat / ((var + self.eps).reshape(shape=[1, -1, 1, 1]).sqrt()) + self.bias.reshape(shape=[1, -1, 1, 1])
        out.name = 'bn_2d_res'
        return out


class Conv1d(Module):
    """1-dimensional convolutional layer.
       
        Args:
            in_channel (int): # channels in input (example: # color channels in image)
            out_channel (int): # channels produced by layer
            kernel_size (int): Edge length of the kernel (i.e. 3x3 kernel <-> kernel_size = 3)
            stride (int): Stride of the convolution (filter)
            padding (tuple, int): Amount of zero-padding applied before the convolution
            weight_initialization (str): Weight initialization method
            with_bias (bool): Performs a convolution with or without a bias
        
        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, in_channel:int, out_channel:int, kernel_size:int, stride:int=1, 
                 padding:Union[tuple, int]=(0, 0), weight_initialization:str="kaiming_normal") -> None:
        super().__init__()
        assert isinstance(padding, (tuple, int)), 'Wrong padding type. Must be integer or tuple of integers'
        if isinstance(padding, (int)): padding = (padding, padding)
        assert len(padding) == 2, 'Wrong padding dimensions. Padding must be an integer of a 2-dimensional tuple'

        self.in_channel, self.out_channel = in_channel, out_channel
        self.stride, self.padding = stride, padding
        self.kernel_size = kernel_size
        self.weight_initialization = weight_initialization

        shape = (self.out_channel, self.in_channel, self.kernel_size)
        self.weight = init_weights(shape, self.weight_initialization, requires_grad=True, is_parameter=True, name="conv_weight_1d")
        self.bias = Tensor.zeros(out_channel, requires_grad=True, is_parameter=True, name="conv_bias_1d")

    def forward(self, x:Tensor) -> Tensor:
        """Forward pass

            Args:
                x (Tensor): (batch_size, in_channel, input_length)
            Returns:
                Tensor: (batch_size, out_channel, output_length)
        """
        x_padded = x.pad1d(self.padding)
        return x_padded.conv1d(self.weight, self.stride) + self.bias.reshape(shape=[1, -1, 1])


class Conv2d(Module):
    r"""
        2-dimensional convolutional layer.

        Args:
            in_channel (int): # channels in input (example: # color channels in image)
            out_channel (int): # channels produced by layer
            kernel_size (tuple): Edge lengths of the kernel
            stride (int): Stride of the convolution (filter)
            padding (int): Padding for the convolution
            weight_initialization (str): Weight initialization method
            with_bias (bool): Performs a convolution with or without a bias
        
        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, in_channel:int, out_channel:int, kernel_size:tuple, stride:tuple=1, 
                 padding:tuple=(0, 0, 0, 0), weight_initialization:str="kaiming_normal") -> None:
        super().__init__()
        assert isinstance(padding, (tuple, int)), 'Wrong padding type. Must be integer or tuple of integers'
        if isinstance(padding, (int)): padding = (padding, padding, padding, padding)
        assert len(padding) == 4, 'Wrong padding dimensions'
        self.in_channel, self.out_channel = in_channel, out_channel
        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        self.stride, self.padding = stride, padding
        self.weight_initialization = weight_initialization

        shape = (self.out_channel, self.in_channel, *self.kernel_size)
        self.weight = init_weights(shape, self.weight_initialization, requires_grad=True, is_parameter=True, name="conv_weight_2d")
        self.bias = Tensor.zeros(self.out_channel, requires_grad=True, is_parameter=True, name="conv_bias_2d")
    
    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, in_channel, width, height)
            Returns:
                Tensor: (batch_size, out_channel, output_dim1, output_dim2)
        """
        x_padded = x.pad2d(self.padding)
        return x_padded.conv2d(self.weight, self.stride) + self.bias.reshape(shape=[1, -1, 1, 1])


class MaxPool1d(Module):
    """Performs a max pooling operation after a 1d convolution
        
        Args:
            pool_size (int): Kernel size
            stride (int): Stride
        
        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, pool_size:int) -> None:
        super().__init__()
        self.pool_size = pool_size
    
    def forward(self, x:Tensor) -> Tensor:
        """Performs a max pooling operation after a 1d convolution

        Args:
            x (Tensor): (batch_size, channel, in_width, in_height)

        Returns:
            Tensor: (batch_size, channel, out_width, out_height)
        """
        out = x.max_pool1d(self.pool_size)
        out.name = 'mpool1d_res'
        return out


class AvgPool1d(Module):
    """Performs an average pooling operation after a 1d convolution
        
        Args:
            pool_size (int): Kernel size
            stride (int): Stride
        
        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, pool_size:int) -> None:
        super().__init__()
        self.pool_size = pool_size
    
    def forward(self, x:Tensor) -> Tensor:
        """Performs an average pooling operation after a 1d convolution

        Args:
            x (Tensor): (batch_size, channel, in_width, in_height)

        Returns:
            Tensor: (batch_size, channel, out_width, out_height)
        """
        out = x.avg_pool1d(self.pool_size)
        out.name = 'avgpool1d_res'
        return out


class MaxPool2d(Module):
    """Performs a max pooling operation after a 2d convolution
    
        Args:
            pool_size (tuple): Kernel size
            stride (int): Stride
        
        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, pool_size:tuple) -> None:
        super().__init__()
        self.pool_size = pool_size if isinstance(pool_size, tuple) else (pool_size, pool_size)
    
    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, channel, in_width, in_height)
            
            Returns:
                Tensor: (batch_size, channel, out_width, out_height)
        """
        out = x.max_pool2d(self.pool_size)
        out.name = 'mpool2d_res'
        return out


class AvgPool2d(Module):
    """Performs an average pooling operation after a 2d convolution

        Args:
            pool_size (tuple): Kernel size
            stride (int): Stride
        
        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, pool_size:tuple) -> None:
        super().__init__()
        self.pool_size = pool_size if isinstance(pool_size, tuple) else (pool_size, pool_size)
    
    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, channel, in_width, in_height)
            
            Returns:
                Tensor: (batch_size, channel, out_width, out_channel)
        """
        out = x.avg_pool2d(self.pool_size)
        out.name = 'avgpool2d_res'
        return out


class Flatten(Module):
    """Layer that flattens all dimensions for each observation in a batch

        >>> x = torch.randn(4, 3, 2) # batch of 4 observations, each sized (3, 2)
        >>> x
        tensor([[[ 0.8816,  0.9773],
                [-0.1246, -0.1373],
                [-0.1889,  1.6222]],

                [[-0.5303,  0.3655],
                [-0.7496,  0.6935],
                [-0.8173,  0.4346]]])
        >>> layer = Flatten()
        >>> out = layer(x)
        >>> out
        tensor([[ 0.8816,  0.9773, -0.1246, -0.1373, -0.1889,  1.6222],
                [-0.5303,  0.3655, -0.7496,  0.6935, -0.8173,  0.4346]])
        >>> out.shape
        torch.size([4, 6]) # batch of 4 observations, each flattened into 1d array size (6,)

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x:Tensor) -> Tensor:
        r"""
            Args:
                x (Tensor): (batch_size, dim_2, dim_3, ...) arbitrary number of dims after batch_size
            Returns:
                out (Tensor): (batch_size, dim_2 * dim_3 * ...) batch_size, then all other dims flattened
        """
        out = x.flatten()
        out.name = 'flatten_res'
        return out


class Dropout(Module):
    """Dropout layer for regularization
    
        Args:
            p (float): Proportion of dropped out connections
        
        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, p:float=0.5) -> None:
        super().__init__()
        self.mask, self.p = None, p
    
    def forward(self, x:Tensor) -> Tensor:
        """Forward pass

            In training mode: Remove connections with probability p
            In validation mode: Multiply by 1-p the output tensor

            Args:
                x (Tensor): Input tensor

            Returns:
                Tensor: Output tensor
        """
        if self.is_train:
            if self.mask is None:
                val = np.random.binomial(1, 1.0 - self.p, size=x.shape)
                self.mask = Tensor(val, name="dropout_mask", is_parameter=True)
            out = x * self.mask
            out.name = 'dropout_res'
            return out
        return x * Tensor(1 - self.p, name='dropout_prob', is_parameter=True)


class NLLLoss(Module):
    """Negative Log-likelihood loss. 
    
       Useful for multi-class classifiation problem.

       The input given through a forward call is expected to contain the log-probabilities
       of each class. To obtain these log-probabilities, simply call the log-softmax method
       on the output vector of your model.
    """
    def __init__(self) -> None:
        pass

    def forward(self, log_probs:Tensor, target:Tensor) -> Tensor:
        """Forward pass

        Args:
            log_probs (Tensor): (batch_size, num_classes)
            target (Tensor): (batch_size, 1)

        Returns:
            Tensor: loss, stored as a float in a Tensor object 
        """
        batch_size, num_classes = log_probs.shape
        labels = target.one_hot(num_classes)
        return -(log_probs * labels).mean()
        

class MSELoss(Module):
    """The MSELoss function.
       Mean squared error function is used for regression problems.

        >>> criterion =  MSELoss()
        >>> criterion(outputs, labels)
        3.241

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self) -> None:
        pass

    def forward(self, predicted:Tensor, target:Tensor) -> Tensor:
        """Forward pass
        
        Args:
            predicted (Tensor): (batch_size, num_classes)
            target (Tensor): (batch_size,)
        Returns:
            Tensor: loss, stored as a float in a tensor
        """
        return ((predicted - target) ** 2).mean()


# ***** Activation functions *****


class ReLU(Module):
    r"""
        ReLU Activation Layer

        Applies a Rectified Linear Unit activation function to 
        the input

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, x:Tensor) -> Tensor:
        """Forward pass.

            ReLU(x) = max(0, x)

            Args:
                x (Tensor): (batch_size, num_features)
            Returns:
                Tensor: (batch_size, num_features)
        """
        return x.relu()


class LeakyReLU(Module):
    r"""
        Leaky ReLU Activation Layer

        Applies a Leaky Rectified Linear Unit activation function to
        the input

        Args:
            alpha (float):

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, alpha=0.01) -> None:
        super().__init__()

        self.alpha = Tensor(alpha, is_parameter=False, name='alpha_relu')

    def forward(self, x:Tensor) -> Tensor:
        """Forward pass

            LeakyReLU(x) = max(alpha * x, x)

            Args:
                x (Tensor): (batch_size, num_features)
            Returns:
                Tensor: (batch_size, num_features)
        """
        return x.relu() - self.alpha * (-x).relu()
    

class Sigmoid(Module):
    r"""
        Sigmoid Activation Layer

        Applies a Sigmoid activation function to the input

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, num_features)
            Returns:
                Tensor: (batch_size, num_features)
        """
        return x.sigmoid()


class Tanh(Module):
    r"""
        Tanh Activation Layer

        Applies a Hyperbolic Tangent activation function to the input

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, num_features)
            Returns:
                Tensor: (batch_size, num_features)
        """
        return x.tanh()

class Swish(Module):
    r"""
        Swish Activation Layer

        Applies a Swish activation function to the input

        Beta is learnt by gradient descent

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self, beta:float=1.0) -> None:
        super().__init__()

        self.beta = Tensor(beta, requires_grad=True, is_parameter=True, name='swish_beta')
    
    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, num_features)
            Returns:
                Tensor: (batch_size, num_features)
        """
        return x * (self.beta * x).sigmoid()


class Mish(Module):
    r"""
        Mish Activation Layer

        Applies a Mish activation function to the input

        Inherits from:
            Module (nn.module.Module)
    """
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x:Tensor) -> Tensor:
        """
            Args:
                x (Tensor): (batch_size, num_features)
            
            Returns:
                Tensor: (batch_size, num_features)
        """
        return x * (1 + x.exp()).log().tanh()


import numpy as np
from nanograd.tensor import Tensor


class Optimizer:
    """Base class for optimizers"""
    def __init__(self, params:list):
        """
            Initializer of the Optimizer class

            Args:
                params (list of Tensors)
        """
        self.params = list(params)
        self.state = []
    
    def step(self):
        """Update rule"""
        raise NotImplementedError("Should be implemented in subclasses of Optimizer")

    def reset(self):
        """Resets all hyperparameters"""
        raise NotImplementedError("Should be implemented in subclasses of Optimizer")

    def zero_grad(self):
        """
            Resets gradients of the parameters 
            to zero. 

            For gradient accumulation, in the training loop, call zero_grad() every
            n (usually 2 or 3) batches
        """
        for param in self.params:
            param.grad = None


class SGD(Optimizer):
    """Implementation of Stochastic Gradient Descent (SGD)
    
        Args:
            params (list of Tensors): parameters to be updated
            lr (float): learning rate
            momentum (float): Nesterov momentum
    """
    def __init__(self, params:list, lr:float=1e-3, momentum:float=0) -> None:
        super(SGD, self).__init__(params)
        assert momentum >= 0.0, "Momentum can't be negative"

        self.lr, self.momentum = lr, momentum
        self.momentums = [Tensor.zeros(p.shape, device=self.params[0].device) for p in self.params]

    def step(self) -> None:
        for i, p in enumerate(self.params):
            self.momentums[i] = self.momentums[i] * self.momentum + self.lr * p.grad
            p.data = (p - self.momentums[i]).data   


class Adam(Optimizer):
    """Implementation of Adam
    
        Args:
            lr (float): learning rate
            beta1 (float): smoothing factor for first moment of the gradient
            beta2 (float): smoothing factor for second moment of the gradient
            eps (float): avoid division by zero
    """
    def __init__(self, params:list, lr:float=1e-3, beta1:float=0.9, 
                 beta2:float=0.999, eps:float=1e-8) -> None:
        super(Adam, self).__init__(params)
        assert (0 <= beta1) and (beta1 < 1), "Smoothing factor must be in [0,1)"
        assert (0 <= beta2) and (beta2 < 1), "Smoothing factor must be in [0,1)"

        self.lr, self.beta1, self.beta2, self.eps = lr, beta1, beta2, eps
        self.t = 0

        self.exp_avg = [Tensor.zeros(p.shape, device=self.params[0].device) for p in self.params]
        self.exp_avg_sq = [Tensor.zeros(p.shape, device=self.params[0].device) for p in self.params]
    
    def step(self) -> None:
        self.t += 1

        for i, p in enumerate(self.params):
            self.exp_avg[i] = self.beta1 * self.exp_avg[i] + (1. - self.beta1) * p.grad
            self.exp_avg_sq[i] = self.beta2 * self.exp_avg_sq[i] + (1. - self.beta2) * (p.grad ** 2)

            bias_correction1 = self.exp_avg[i] / (1. - (self.beta1 ** self.t))    
            bias_correction2 = self.exp_avg_sq[i] / (1. - (self.beta2 ** self.t))

            p.data = (p - self.lr * bias_correction1 / (bias_correction2.sqrt() + self.eps)).data


class AdamW(Optimizer):
    """Implementation of AdamW: Adam with enhanced weight decay

        Args:
            lr (float): learning rate
            weight_decay (float): L2-regularization parameter
            beta1 (float): smoothing factor for first moment of the gradient
            beta2 (float): smoothing factor for second moment of the gradient
            eps (float): avoid division by zero
    """
    def __init__(self, params:list, lr:float=1e-3, weight_decay:float=1e-2, 
                 beta1:float=0.9, beta2:float=0.999, eps:float=1e-8) -> None:
        super(AdamW, self).__init__(params)
        assert (0 <= beta1) and (beta1 < 1), "Smoothing factor must be in [0,1)"
        assert (0 <= beta2) and (beta2 < 1), "Smoothing factor must be in [0,1)"

        self.lr, self.weight_decay, self.eps = lr, weight_decay, eps
        self.beta1, self.beta2 = beta1, beta2
        self.t = 0

        self.exp_avg = [Tensor.zeros(p.shape, device=self.params[0].device) for p in self.params]
        self.exp_avg_sq = [Tensor.zeros(p.shape, device=self.params[0].device) for p in self.params]        

    def step(self):
        self.t += 1

        for i, p in enumerate(self.params):
            bias_correction1 = 1 - self.beta1 ** self.t
            bias_correction2 = 1 - self.beta2 ** self.t

            self.exp_avg[i] = self.beta1 * self.exp_avg[i] + (1. - self.beta1) * p.grad
            self.exp_avg_sq[i] = self.beta2 * self.exp_avg_sq[i] + (1. - self.beta2) * (p.grad ** 2)

            denom = (self.exp_avg_sq[i] / bias_correction2).sqrt() + self.eps
            step_size = self.lr / bias_correction1

            p.data = (p * (1 - self.lr * self.weight_decay) - step_size * (self.exp_avg[i] / denom)).data


from enum import Enum


class Device(Enum):
    """
        Enumeration of the devices supported by
        Nanograd. 
        Currently, Nanograd only supports CPU and GPU.
    """
    CPU = 1
    GPU = 2


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="nanograd",
    version="1.0.4",
    description="A lightweight deep learning framework",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/PABannier/nanograd",
    author="PAB",
    author_email="pierre-antoine.bannier@polytechnique.edu",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(exclude=('tests', 'examples', 'utils')),
    include_package_data=True,
    install_requires=[
        "numpy", 
        "matplotlib", 
        "graphviz", 
        "pyopencl", 
        "mako"
    ],
)


from nanograd import tensor
from nanograd.nn import ops_gpu
from nanograd.device import Device


def backward(grad_fn, grad_of_outputs):
    """
        Recursive DFS that traverses comp graph, handing back gradients as it goes.

        Args:
            grad_fn (BackwardFunction or AccumulateGrad): Current node type from
                                                        parent's `.next_functions`
            grad_of_output (Tensor): Gradient of the final node w.r.t. current output

        Returns:
            No return statement needed.
    """

    if grad_fn:
        gradients = grad_fn.apply(grad_of_outputs)
        functions = grad_fn.next_functions
        
        for i in range(len(functions)):
            if functions[i]:
                backward(functions[i], gradients[i])

class Function:
    """
        Superclass for linking nodes to the computational graph.
        Operations in `functional.py` should inherit from this
    """
    @staticmethod
    def forward(ctx, *args):
        raise NotImplementedError("All subclasses must implement forward")

    @staticmethod
    def backward(ctx, *grad_outputs):
        raise NotImplementedError("All subclasses must implement backward")

    @classmethod
    def apply(cls, *args, **kwargs):
        """
            Runs forward of subclass and links node to the comp graph.

            Args:
                cls (subclass of Function): Current function, such as Add, Sub, etc.
                args (tuple): arguments for the subclass's `.forward()`.

            Returns:
                Tensor: Output tensor from operation that stores the current node.
        """
        # Creates BackwardFunction obj representing the current node
        backward_function = BackwardFunction(cls)

        cl_ctx, cl_queue = None, None

        # Gets OpenCL objects
        try:
            cl_ctx, cl_queue = kwargs['cl_ctx'], kwargs['cl_queue']
            backward_function.ctx.cl_ctx = cl_ctx
            backward_function.ctx.cl_queue = cl_queue
        except KeyError:
            pass

        # Run subclass's forward with context manager and operation input args
        output_tensor = cls.forward(backward_function.ctx, *args)

        # 1) For each parent tensor in args, add their node to `backward_function.next_functions`
        for arg in args:
            if isinstance(arg, tensor.Tensor):
                if not arg.grad_fn:
                    if arg.requires_grad and arg.is_leaf:
                        arg.grad_fn = AccumulateGrad(arg, cl_ctx, cl_queue)
                    elif arg.requires_grad and not arg.is_leaf:
                        arg.grad_fn = BackwardFunction(cls)
                    elif not arg.requires_grad and arg.is_leaf:
                        arg.grad_fn = None
                    else:
                        raise Exception("Gradient-disabled nodes are necessarily leaves")
                
                backward_function.next_functions.append(arg.grad_fn)
            else:
                backward_function.next_functions.append(None)

        # 2) Store current node in output tensor (see `tensor.py` for ideas)
        output_tensor.grad_fn = backward_function

        return output_tensor


class AccumulateGrad:
    """
        Represents node where gradient must be accumulated.

        Args:
            tensor (Tensor): The tensor where the gradients are accumulated in `.grad`
    """
    def __init__(self, tensor, cl_ctx=None, cl_queue=None):
        self.variable = tensor
        self.next_functions = [] 
        self.function_name = "AccumulateGrad"

        self.cl_ctx, self.cl_queue = cl_ctx, cl_queue

    def apply(self, arg):
        """
            Accumulates gradient provided.

            Args:
                arg (Tensor): Gradient to accumulate
        """
        if self.variable.grad is None:
            self.variable.grad = tensor.Tensor(arg.data, device=self.variable.device)
        else:
            if self.variable.device == Device.GPU:
                self.variable.grad.data = ops_gpu.element_wise_binary_op(
                    self.cl_ctx, self.cl_queue, 'a+b', arg.data, self.variable.grad.data)
            else:
                self.variable.grad.data += arg.data


class ContextManager:
    """
        Used to pass variables between a function's `.forward()` and `.backward()`.
        (Argument "ctx" in these functions)

        To store a tensor:
        >>> ctx.save_for_backward(<tensors>, <to>, <store>)

        To store other variables (like integers):
        >>> ctx.<some_name> = <some_variable>
    """
    def __init__(self):
        self.saved_tensors = [] 

    def save_for_backward(self, *args):
        r"""
            Saves TENSORS only

            Args:
                args (Tensor(s)): Tensors to store
        """
        for arg in args:
            # Raises error if arg is not tensor (i warned you)
            if type(arg).__name__ != "Tensor":
                raise Exception("Got type {} of object {}. \nOnly Tensors should be saved in save_for_backward. For saving constants, just save directly as a new attribute.".format(type(arg), arg))

            self.saved_tensors.append(arg.copy())


class BackwardFunction:
    """
        Representing an intermediate node where gradient must be passed.
        Stored on output tensor of operation during `Function.apply()`

        Args:
            cls (subclass of Function): Operation being run. Don't worry about this;
                                        already handled in `Function.apply()`
    """
    def __init__(self, cls):
        self.ctx = ContextManager()
        self._forward_cls = cls

        # Nodes of parents, populated in `Function.apply`
        self.next_functions = []

        # The name of the operation as a string (for convenience)
        self.function_name = "BackwardFunction"

    def apply(self, *args):
        """
            Generates gradient by running the operation's `.backward()`.

            Args:
                args: Args for the operation's `.backward()`

            Returns:
                Tensor: gradient of parent's output w.r.t. current output
        """
        return self._forward_cls.backward(self.ctx, *args)



import numpy as np
from nanograd.nn.conv_ops import get_conv1d_output_size, get_conv2d_output_size

# *************************************
# ************** Helpers **************
# *************************************

def sigmoid(x:np.ndarray) -> np.ndarray:
    return 1 / (1 + np.exp(-x))

def unbroadcast(grad:np.ndarray, shape:tuple, to_keep:int=0) -> np.ndarray:
    while len(grad.shape) != len(shape):
        grad = grad.sum(axis=0)
    for i in range(len(shape) - to_keep):
        if grad.shape[i] != shape[i]:
            grad = grad.sum(axis=i, keepdims=True)
    return grad

def inner_slice(a, indices):
    """
        Helper function to slice a Tensor

        Args:
            a (np.ndarray): array to slice
            indices (list): list of indices 
        
        ..note: Length must match the number of dimensions of x
    """
    padding = [(max(0, -p[0]), max(0, p[1]-a.shape[i])) for i, p in enumerate(indices)]
    a = np.pad(a, padding, mode="constant")
    slices = [(p[0]+padding[i][0], p[1]+padding[i][0]) for i, p in enumerate(indices)]
    return a[tuple([slice(x[0], x[1], None) for x in slices])]

def one_hot_encoding(a, num_classes):
    idx = a.astype(int)
    out = np.zeros((idx.shape[0], num_classes))
    out[np.arange(len(out)), idx] = 1
    return out

# *************************************
# *********** Forward passes **********
# *************************************

def unsqueeze_forward(a, axis):
    return np.expand_dims(a, axis)

def squeeze_forward(a, axis):
    return np.squeeze(a, axis)

def slice_forward(a, indices):
    return inner_slice(a, indices)

def transpose_forward(a):
    return a.T

def reshape_forward(a, shape):
    return a.reshape(shape)

def max_forward(a, axis):
    out = np.amax(a, axis=None if axis is None else tuple(axis), keepdims=True) 
    if axis is not None:
        out = out.reshape([a.shape[i] for i in range(len(a.shape)) if i not in axis])
    return out

def min_forward(a, axis):
    out = np.amin(a, axis=None if axis is None else tuple(axis), keepdims=True)
    if axis is not None:
        out = out.reshape([a.shape[i] for i in range(len(a.shape)) if i not in axis])
    return out

def sum_forward(a, axis):
    if axis is None:
        return np.array(a.sum())
    return a.sum(axis=axis)

def add_forward(a, b):
    return a + b

def mul_forward(a, b):
    return a * b

def matmul_forward(a, b):
    return a @ b

def log_forward(a):
    return np.log(a)

def exp_forward(a):
    return np.exp(a)

def neg_forward(a):
    return -a

def pow_forward(a, exp):
    return a ** exp

def relu_forward(a):
    return np.maximum(a, 0)

def sigmoid_forward(a):
    return 1.0 / (1.0 + np.exp(-a))

def tanh_forward(a):
    return np.tanh(a)

def conv1d_forward(a, weight, stride):
    batch_size, in_channel, signal_length = a.shape
    num_filters, _, kernel_length = weight.shape
    output_length = get_conv1d_output_size(signal_length, kernel_length, stride, 0)

    stride_shape = (signal_length, 1, in_channel * signal_length, stride)
    strides = a.data.itemsize * np.array(stride_shape)

    cols = np.lib.stride_tricks.as_strided(
        x=a,
        strides=strides,
        shape=(in_channel, kernel_length, batch_size, output_length),
        writeable=False
    )

    cols = cols.transpose(2, 0, 1, 3)
    weight = weight.transpose(1, 2, 0) 

    ret = np.tensordot(cols, weight, axes=[(1, 2), (0, 1)])
    ret = ret.transpose(0, 2, 1)
    return ret, cols.transpose(0, 1, 3, 2) 


def conv2d_forward(a, weight, stride):
    batch_size, in_channel, im_height, im_width = a.shape
    num_filters, _, kernel_height, kernel_width = weight.shape
    output_height, output_width = get_conv2d_output_size(im_height, im_width, (kernel_height, kernel_width), stride, 0)

    strides = (im_height * im_width, im_width, 1, in_channel * im_height * im_height, stride * im_width, stride)
    strides = a.itemsize * np.array(strides)

    cols = np.lib.stride_tricks.as_strided(
        x=a,
        shape=(in_channel, kernel_height, kernel_width, batch_size, output_height, output_width),
        strides=strides,
        writeable=False
    )

    cols = cols.transpose(3, 0, 1, 2, 4, 5)
    weight = weight.transpose(1, 2, 3, 0)

    #jiyxYX,iyxk -> jYXk -> jkYX
    ret = np.tensordot(cols, weight, axes=[(1, 2, 3), (0, 1, 2)])
    ret = ret.transpose(0, 3, 1, 2)
    return ret, cols.transpose(0, 1, 4, 5, 2, 3) 

# *************************************
# ********** Backward passes **********
# *************************************

def unsqueeze_backward(grad_output, axis):
    return grad_output.squeeze(axis)

def squeeze_backward(grad_output, axis):
    return np.expand_dims(grad_output, axis)

def add_backward(grad_output, a_shape, b_shape):
    grad_a = np.ones(a_shape) * grad_output.data
    grad_b = np.ones(b_shape) * grad_output.data
    return unbroadcast(grad_a, a_shape), unbroadcast(grad_b, b_shape)

def mul_backward(grad_output, a, b):
    grad_a = grad_output * b
    grad_b = grad_output * a
    return unbroadcast(grad_a, a.shape), unbroadcast(grad_b, b.shape)

def matmul_backward(grad_output, a, b):
    grad_a = np.matmul(grad_output, b.T)
    grad_b = np.matmul(a.T, grad_output)
    return grad_a, grad_b

def log_backward(grad_output, a):
    return grad_output / a

def exp_backward(grad_output, a):
    return grad_output * np.exp(a)

def neg_backward(grad_output):
    return -grad_output

def pow_backward(grad_output, a, exp):
    return exp * (a ** (exp-1)) * grad_output

def relu_backward(grad_output, a):
    return grad_output * (a >= 0)

def sigmoid_backward(grad_output, a):
    return grad_output * sigmoid(a) * (1 - sigmoid(a))

def tanh_backward(grad_output, a):
    return grad_output * (1 - np.power(np.tanh(a), 2))

def slice_backward(grad_output, shape, fwd_indices):
    indices = [(0 - p[0], grad_output.shape[i] + (shape[i] - p[1])) for i, p in enumerate(fwd_indices)]
    return inner_slice(grad_output, indices)

def transpose_backward(grad_output):
    return grad_output.T

def reshape_backward(grad_output, shape):
    return grad_output.reshape(shape)

def max_backward(grad_output, inp, out, axis):
    shape = [1 if axis is None or i in axis else inp.shape[i] for i in range(len(inp.shape))]
    ret2 = (inp == out.reshape(shape))
    div = ret2.sum(axis=None if axis is None else tuple(axis), keepdims=True) 
    return ret2 * (grad_output.reshape(shape)).data / div

def min_backward(grad_output, inp, out, axis):
    shape = [1 if axis is None or i in axis else inp.shape[i] for i in range(len(inp.shape))]
    ret2 = (inp == out.reshape(shape))
    div = ret2.sum(axis=None if axis is None else tuple(axis), keepdims=True) 
    return ret2 * (grad_output.reshape(shape)).data / div

def sum_backward(grad_output, a, axis):
    axis = [axis] if type(axis) == int else axis
    shape = [1 if axis is None or i in axis else a.shape[i] for i in range(len(a.shape))]
    return grad_output.reshape(shape) + np.zeros_like(a) # Useful for broadcasting

def conv1d_backward(grad_output, x, x_reshaped, weight, stride):
    batch_size, in_channel, signal_length = x.shape
    num_filters, _, kernel_length = weight.shape
    _, _, output_length = grad_output.shape

    #grad_weight = np.einsum('ikX, ijXx -> kjx', grad_output, x_reshaped) SLOWER than using tensordot
    grad_weight = np.tensordot(grad_output, x_reshaped, axes=[(0, 2), (0, 2)])

    grad_x = np.zeros((batch_size, in_channel, signal_length), dtype=grad_output.dtype)

    for k in range(output_length):
        X = k % output_length
        iX = X * stride

        #grad_x[:, :, iX:iX+kernel_length] += np.einsum('ik, kjy->ijy', grad_output[:, :, X], weight) SLOWER than using tensordot
        grad_x[:, :, iX:iX+kernel_length] += np.tensordot(grad_output[:, :, X], weight, axes=[(1), (0)])
    
    grad_x = grad_x.reshape((batch_size, in_channel, signal_length))

    return grad_x, grad_weight
        
def conv2d_backward(grad_output, x, x_reshaped, weight, stride):
    batch_size, in_channel, im_height, im_width = x.shape
    num_filters, _, kernel_height, kernel_width = weight.shape
    _, _, output_height, output_width = grad_output.shape
    
    #grad_weight = np.einsum('ikYX, ijYXyx -> kjyx', grad_output, x_reshaped) SLOWER than using tensordot 
    grad_weight = np.tensordot(grad_output, x_reshaped, axes=[(0,2,3),(0,2,3)])

    grad_x = np.zeros((batch_size, in_channel, im_height, im_width), dtype=grad_output.dtype)

    for k in range(output_height * output_width):
        X, Y = k % output_width, k // output_width
        iX, iY = X * stride, Y * stride

        # grad_x[:,:, iY:iY+kernel_height, iX:iX+kernel_width] += np.einsum('ik,kjyx->ijyx', grad_output[:,:,Y,X], weight) 
        # SLOWER than using tensordot
        grad_x[:,:, iY:iY+kernel_height, iX:iX+kernel_width] += np.tensordot(grad_output[:,:,Y,X], weight, axes=[(1), (0)])

    grad_x = grad_x.reshape((batch_size, in_channel, im_height, im_width))

    return grad_x, grad_weight


from typing import Union
import warnings

import numpy as np
from nanograd.viz.comp_graph import ForwardGraphVisualizer, BackwardGraphVisualizer

from nanograd.device import Device
from nanograd import autograd_engine
from nanograd.autograd_engine import Function
from nanograd.nn.ops_gpu import GPUBuffer

import nanograd.nn.ops_cpu as ops_cpu
import nanograd.nn.ops_gpu as ops_gpu
from nanograd.nn.conv_ops import (get_conv1d_output_size, get_conv2d_output_size, 
                                  get_im2col_indices, col2im)

try:
    import pyopencl as cl
    PYOPENCL_AVAILABLE = True

except ImportError:
    PYOPENCL_AVAILABLE = False
    warnings.warn("PyOpenCL is not available on this computer. Can't use \
                   parallel computing. Please install it to move comptutations \
                   to move the GPU.")


cl_ctx, cl_queue = None, None

def get_gpu_context_and_queue():
    """
        If GPU is enabled, get_gpu_context_and_queue populates global variables
        cl_ctx and cl_queue used for GPU computations. 
        
        cl_ctx and cl_queue are PyOpenCL objects used for parallelized computations 
        on the GPU. More precisely, cl_ctx is expected to be an instance of PyOpenCL 
        Context and cl_queue is expected to be an instance of PyOpenCL CommandQueue.
    """
    global cl_ctx, cl_queue
    devices = cl.get_platforms()[0].get_devices(device_type=cl.device_type.GPU)
    if len(devices) == 0:
        devices = cl.get_platforms()[0].get_devices(device_type=cl.device_type.CPU)
    cl_ctx = cl.Context(devices=devices)
    cl_queue = cl.CommandQueue(cl_ctx)


class Tensor:
    """
        Tensor is the basic operator object of Nanograd. It is 
        a wrapper class around NumPy array. 

        Args:
            data (np.ndarray or int or float or GPUBuffer): Contains data to be stored. It can 
                be a scalar, a vector, a matrix or a tensor (multi-dimensional array).

            requires_grad (bool, optional): If ``True``, a gradient will be stored and 
                accumulated as a property of an object. If ``False``, the grad property 
                remains None. No gradient is stored in the object.

            is_leaf (bool, optional): If ``True``, the corresponding node in the computational
                graph does not have any parents. Note that it is usually the case for parameters
                in a neural network. If ``False``, the node is expected to have parents. 
                Usually a non-leaf node results from a basic operation.

            is_parameter (bool, optional): the Tensor contains trainable parameters. It is useful
                when building a neural network.

            device (Device, optional): the device to use for computations. Currently, Nanograd only
                supports CPU and GPU.

            name (str, optional): A name for the Tensor for visualization purposes.

            op (str, optional): The operation used to create the tensor. Could be an addition, a substraction,
                a product.

        ..note::
            If the node is gradient-enabled, the grad property is populated with a gradient Tensor
            object during backpropagation.
    """
    def __init__(self, 
                 data:Union[np.ndarray, GPUBuffer], 
                 requires_grad:bool=False, 
                 is_leaf:bool=True, 
                 is_parameter:bool=False,
                 device:Device=Device.CPU,
                 name:str='no_name',
                 op:str=None) -> None:
        self.data = self._move_data(data, device)
        self.requires_grad, self.is_leaf = requires_grad, is_leaf
        self.is_parameter = is_parameter

        self.device = device

        self.grad, self.grad_fn = None, None
        self.name, self.op = name, op
        self.children = []

        if not self.requires_grad and not self.is_leaf:
            raise Exception("A non-leaf node must be gradient-enabled")
    
    @property
    def shape(self) -> tuple:
        return self.data.shape
    
    @property
    def dtype(self) -> np.dtype:
        return self.data.dtype
    
    # ****************************************
    # ***** Class methods / Initializers *****
    # ****************************************

    @classmethod
    def zeros(cls, *shape, **kwargs):
        """Creates a Tensor filled with zeros"""
        return cls(np.zeros(*shape, dtype=np.float32), **kwargs)
    
    @classmethod
    def ones(cls, *shape, **kwargs):
        """Creates a Tensor filled with ones"""
        return cls(np.ones(*shape, dtype=np.float32), **kwargs)
    
    @classmethod
    def arange(cls, *interval, **kwargs):
        """Creates a Tensor filled with values in range"""
        return cls(np.arange(*interval).astype(np.float32), **kwargs)
    
    @classmethod
    def randn(cls, *shape, **kwargs):
        """
            Creates a Tensor filled with values drawn from the
            standard Gaussian distribution
        """
        return cls(np.random.randn(*shape).astype(np.float32), **kwargs)
    
    @classmethod
    def normal(cls, loc, scale, *shape, **kwargs):
        """
            Creates a Tensor filled with values drawn from the
            a custom Gaussian distribution
        """
        return cls(np.random.normal(loc, scale, *shape).astype(np.float32), **kwargs)

    @classmethod
    def randint(cls, low, high, *shape, **kwargs):
        """
            Creates a Tensor filled with integer values drawn in the
            range between low and high
        """
        return cls(np.random.randint(low, high, *shape).astype(np.float32), **kwargs)
    
    @classmethod
    def eye(cls, dim, **kwargs):
        """
            Creates a 2-dimensional Tensor (matrix) equal to the
            identity matrix.
        """
        return cls(np.eye(dim).astype(np.float32), **kwargs)
    
    # ****************************************
    # *********** CPU/GPU support ************
    # ****************************************

    @staticmethod
    def _move_data(data:Union[np.ndarray, GPUBuffer], device:Device):
        """
            Moves data to the device specified

            Args:
                data (np.ndarray or GPUBuffer): data to be moved
                device (Device): the destination device
            
            Returns:
                data (np.ndarray or GPUBuffer): numpy array if CPU else cl.Buffer
        """
        assert device in Device, "Unsupported device. Only CPU and GPU available."

        if isinstance(data, GPUBuffer):
            if device == Device.GPU:
                return data
            cpu_data = np.empty(data.shape, dtype=np.float32)
            cl.enqueue_copy(cl_queue, cpu_data, data.cl, is_blocking=True)
            return cpu_data
        
        if not isinstance(data, np.ndarray):
            data = np.array(data, dtype=np.float32)

        if device == Device.GPU:
            if cl_ctx is None:
                get_gpu_context_and_queue()
            return GPUBuffer(cl_ctx, data.shape, hostbuf=data)
        return data

    def to(self, device:Device):
        """
            Moves the Tensor to the specified device 
            
            Args:
                device (Device): the destination device
        """
        self.data, self.device = self._move_data(self.data, device), device
        if self.grad:
            self.grad.to(device) # Recursive call to move the gradient to specified device

    def cpu(self):
        self.to(Device.CPU)
        return self
    
    def gpu(self):
        if not PYOPENCL_AVAILABLE:
            raise Exception("OpenCL is not installed in this environment. Please consider running \
                             pip install pyopencl to benefit from GPU-accelerated computations.")
        self.to(Device.GPU)
        return self
        

    # ****************************************
    # *************** Backprop ***************
    # ****************************************

    def backward(self) -> None:
        """
            Initiates backpropagation.

            ..note::
                The gradient of the loss with respect to the parameters is a multi-dimensional array filled
                with ones.
        """
        if not self.requires_grad:
            raise Exception("Can't initiate backprop from gradient-disabled tensor")
        autograd_engine.backward(self.grad_fn, Tensor.ones(self.shape, device=self.device))
    
    def copy(self):
        """Creates a copy of the tensor"""
        return Tensor(self.data, device=self.device)

    # ****************************************
    # *********** Magic functions ************
    # ****************************************
    
    def __str__(self):
        return f"<NanoTensor({str(self.data)}, " + \
               f"grad_fn={self.grad_fn.__class__.__name__  if self.grad_fn else None}), " + \
               f"name={self.name}," + \
               f"device={self.device}>"
    
    def __repr__(self):
        return self.__str__()

    def __getitem__(self, args):
        if args is None: 
            args = []
        elif type(args) in [list, tuple]: 
            pass
        else: 
            args = [args]

        indices = []

        for i, arg in enumerate(args):
            start, stop = arg.start, arg.stop

            if start is None:
                start = 0
            elif not np.issubdtype(type(start), int):
                raise TypeError(f"Indices must be integer. Got {type(start)}")
            
            if stop is None:
                stop = self.shape[i]
            elif not np.issubdtype(type(stop), int):
                raise TypeError(f"Indices must be integer. Got {type(stop)}")
            elif stop < 0:
                stop = self.shape[i] + stop

            assert arg.step is None or arg.step == 1, "Custom step not yet implemented"
            indices.append((start, stop))
        
        indices += [(0, shape[i]) for i in range(len(args), len(self.shape))]
        
        return Slice.apply(self, indices, cl_ctx=cl_ctx, cl_queue=cl_queue)

    # ****************************************
    # ********** Basic operations ************
    # ****************************************
        
    def __add__(self, other):
        if isinstance(other, (int, float)):
            other = Tensor(other)
        return Add.apply(self, other, cl_ctx=cl_ctx, cl_queue=cl_queue)

    def __radd__(self, other):
        return self + other
    
    def __neg__(self):
        return Neg.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def __sub__(self, other):
        return self + (- other)
    
    def __rsub__(self, other):
        return -self + other

    def __pow__(self, exp):
        return Pow.apply(self, exp, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def __mul__(self, other):
        if isinstance(other, (int, float)):
            other = Tensor(other)
        return Mul.apply(self, other, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def __rmul__(self, other):
        return self * other
    
    def __truediv__(self, other):
        return self * (other ** -1.0)
    
    def __matmul__(self, other):
        return MatMul.apply(self, other, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    # ****************************************
    # ******** Advanced operations ***********
    # ****************************************
    
    def sum(self, axis=None):
        return Sum.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def mean(self, axis=None, keepdims:bool=False):
        out = self.sum(axis=axis)
        coeff = np.prod(out.shape) / np.prod(self.shape)
        return out * Tensor(coeff, device=self.device)
    
    def reshape(self, shape:tuple):
        return Reshape.apply(self, shape, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def T(self):
        return Transpose.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)

    def max(self, axis=None):
        return Max.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def min(self, axis=None):
        return Min.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def log(self):
        return Log.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def exp(self):
        return Exp.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)

    def sqrt(self):
        return self ** (1/2)

    # ****************************************
    # ****** Miscellaneous operations ********
    # ****************************************
    
    def one_hot(self, num_classes):
        return OneHot.apply(self, num_classes, cl_ctx=cl_ctx, cl_queue=cl_queue)

    def unsqueeze(self, axis):
        return Unsqueeze.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def squeeze(self, axis):
        return Squeeze.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)

    def flatten(self):
        dim1, dim2 = self.shape[0], np.prod(self.shape[1:])
        out = self.reshape((dim1, dim2))
        return out

    # ****************************************
    # ******** Activation functions **********
    # ****************************************

    def relu(self):
        return ReLU.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def sigmoid(self):
        return Sigmoid.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)

    def tanh(self):
        return Tanh.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def log_softmax(self):
        batch_size, num_classes = self.shape
        a = self.max(axis=1).reshape((batch_size, 1)) # Log-exp trick
        out = self - a - (self - a).exp().sum(axis=1).unsqueeze(0).log().reshape((batch_size, 1))
        return out

    # ****************************************
    # ********* Conv/Pool operations *********
    # ****************************************

    def _pool1d(self, field_length:int):
        """1-dimensional pooling operation

        Args:
            field_length (int): length of the pooling kernel
        
        Returns:
            x_reshaped (Tensor): pooled Tensor
        """
        x_unpadded = self[:, :, :self.shape[2] - self.shape[2] % field_length]

        shape = (x_unpadded.shape[0], x_unpadded.shape[1], 
                 x_unpadded.shape[2] // field_length, field_length)

        x_reshaped = x_unpadded.reshape(shape=shape)
        return x_reshaped 
    
    def max_pool1d(self, kernel_size:int=2):
        """MaxPooling1d operation

        Args:
            kernel_size (int, optional): Kernel length for pooling operation. Defaults to 2.
        """
        return self._pool1d(kernel_size).max(axis=3)
    
    def avg_pool1d(self, kernel_size:int=2):
        """AvgPooling1d operation

        Args:
            kernel_size (int, optional): Kernel length for pooling operation. Defaults to 2.
        """
        return self._pool1d(kernel_size).mean(axis=3)
        
    def _pool2d(self, pool_size:tuple):
        """
            2-dimensional pooling operation

            Args:
                pool_size (tuple): height of the pooling kernel
            
            Returns:
                x_reshaped (Tensor): pooled Tensor
        """
        x_unpadded = self[:, :, 
            :self.shape[2] - self.shape[2] % pool_size[0], 
            :self.shape[3] - self.shape[3] % pool_size[1]]
        
        shape = (x_unpadded.shape[0], x_unpadded.shape[1], 
                 x_unpadded.shape[2] // pool_size[0], pool_size[0], 
                 x_unpadded.shape[3] // pool_size[1], pool_size[1])

        x_reshaped = x_unpadded.reshape(shape=shape)

        return x_reshaped
    
    def max_pool2d(self, pool_size:tuple=(2, 2)):
        """MaxPooling2d operation
        
            Args:
                kernel_size (tuple): Kernel length for pooling operation
        """
        return self._pool2d(pool_size).max(axis=(3, 5))
    
    def avg_pool2d(self, pool_size:tuple=(2, 2)):
        """AvgPooling2d operation
        
            Args:
                kernel_size (tuple): Kernel length for pooling operation
        """
        return self._pool2d(pool_size).mean(axis=(3, 5))

    def pad1d(self, pad:tuple):
        """Padding for one-dimensional signal

        Args:
            pad (tuple): Amount of padding before and after the signal

        Returns:
            Tensor: Padded signal
        """
        return self[:, :, -pad[0]:int(self.shape[2])+pad[1]]

    def pad2d(self, pad:tuple):
        """Padding for two-dimensional images

        Args:
            pad (tuple): 4-dimensional tuple. Amount of padding to be applied before and
                         after the signal along 2 dimensions

        Returns:
            Tensor: Padded signal
        """
        return self[:, :, -pad[2]:int(self.shape[2])+pad[3], -pad[0]:int(self.shape[3])+pad[1]]

    def conv1d(self, weight, stride:int):
        """1d convolution
        
        Args:
            weight (Tensor): Filter weight (out_channel, in_channel, kernel_length)
            stride (int): Stride of the convolution operation
        """
        return Conv1d.apply(self, weight, stride, cl_ctx=cl_ctx, cl_queue=cl_queue)
    
    def conv2d(self, weight, stride):
        """2d convolution
        
        Args:
            weight (Tensor): Filter weight (out_channel, in_channel, *kernel_size)
            stride (int): Stride of the convolution operation
        """
        return Conv2d.apply(self, weight, stride, cl_ctx=cl_ctx, cl_queue=cl_queue)

    # ****************************************
    # ************ Visualization *************
    # ****************************************

    def plot_forward(self, rankdir="LR"):
        r"""
            Plots a forward computational graph

            Args:
                rankdir (str): LR (left to right) and TB (top to bottom)
        """
        visualizer = ForwardGraphVisualizer()
        return visualizer.visualize(self, rankdir=rankdir)
    
    def plot_backward(self, rankdir="LR"):
        r"""
            Plots a backward computational graph

            Args:
                rankdir (str): LR (left to right) and TB (top to bottom)
        """
        visualizer = BackwardGraphVisualizer()
        return visualizer.visualize(self, rankdir=rankdir)


# ****************************************
# ************** Functional **************
# ****************************************

class OneHot(Function):
    @staticmethod
    def forward(ctx, a, num_classes):
        requires_grad = a.requires_grad
        is_leaf = not a.requires_grad

        if a.device == Device.CPU:
            out = ops_cpu.one_hot_encoding(a.data, num_classes)
        else:
            out = ops_gpu.one_hot_encoding(ctx.cl_ctx, ctx.cl_queue, a.data, num_classes)
        
        return Tensor(out, device=a.device, requires_grad=requires_grad,
                      is_leaf=is_leaf)
    
    @staticmethod
    def backward(ctx, grad_output):
        # No backward pass since one-hot encoding is applied to a target
        # tensor whose gradient is None
        return None


class Unsqueeze(Function):
    @staticmethod
    def forward(ctx, a, axis):
        requires_grad = a.requires_grad
        is_leaf = not a.requires_grad

        ctx.axis = axis

        if a.device == Device.CPU:
            out = ops_cpu.unsqueeze_forward(a.data, axis)
        else:
            out = ops_gpu.unsqueeze_forward(ctx.cl_ctx, ctx.cl_queue, a, axis)

        return Tensor(out, device=a.device, requires_grad=requires_grad,
                      is_leaf=is_leaf)
    
    @staticmethod
    def backward(ctx, grad_output):
        axis = ctx.axis

        if grad_output.device == Device.CPU:
            grad = ops_cpu.unsqueeze_backward(grad_output.data, axis)
        else:
            grad = ops_gpu.unsqueeze_backward(ctx.cl_ctx, ctx.cl_queue, grad_output, axis)
        
        return Tensor(grad, device=grad_output.device), None


class Squeeze(Function):
    @staticmethod
    def forward(ctx, a, axis):
        is_squeezed = False
        requires_grad = a.requires_grad
        is_leaf = not a.requires_grad

        if a.shape[axis] == 1: # If dimension > 1, can't squeeze
            is_squeezed = True
            if a.device == Device.CPU:
                out = ops_cpu.squeeze_forward(a.data, axis)
            else:
                out = ops_gpu.squeeze_forward(ctx.cl_ctx, ctx.cl_queue, a, axis)
        else:
            is_squeezed = False
            out = a.data

        ctx.axis = axis
        ctx.is_squeezed = is_squeezed
        
        return Tensor(out, device=a.device, requires_grad=requires_grad,
                      is_leaf=is_leaf)
    
    @staticmethod
    def backward(ctx, grad_output):
        axis, is_squeezed = ctx.axis, ctx.is_squeezed

        if is_squeezed:
            if grad_output.device == Device.CPU:
                grad = ops_cpu.squeeze_backward(grad_output.data, axis)
            else:
                grad = ops_gpu.squeeze_backward(ctx.cl_ctx, ctx.cl_queue, grad_output, axis)
        else:
            grad = grad_output.data
        
        return Tensor(grad, device=grad_output.device), None


class Slice(Function):
    @staticmethod
    def forward(ctx, a, indices=None):
        ctx.shape, ctx.indices = a.shape, indices

        requires_grad = a.requires_grad
        is_leaf = not a.requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.slice_forward(a.data, indices)
        else:
            out_data = ops_gpu.slice_forward(ctx.cl_ctx, ctx.cl_queue, a.data, indices)
        
        out = Tensor(out_data, requires_grad=requires_grad, 
                            is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'slice'

        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        shape, fwd_indices = ctx.shape, ctx.indices

        if grad_output.device == Device.CPU:
            grad = ops_cpu.slice_backward(grad_output.data, shape, fwd_indices)
        else:
            grad = ops_gpu.slice_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, shape, fwd_indices)
        
        return Tensor(grad, device=grad_output.device), None


class Transpose(Function):
    @staticmethod
    def forward(ctx, a):
        if len(a.shape) > 2:
            raise Exception("Arg for Transpose must be 1D or 2D tensor: {}".format(a.shape))
        
        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.transpose_forward(a.data)
        else:
            out_data = ops_gpu.transpose_forward(ctx.cl_ctx, ctx.cl_queue, a.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                     is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'transpose'
        return out

    @staticmethod
    def backward(ctx, grad_output):

        if grad_output.device == Device.CPU:
            grad = ops_cpu.transpose_backward(grad_output.data)
        else:
            grad = ops_gpu.transpose_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data)

        return Tensor(grad, device=grad_output.device), None


class Reshape(Function):
    @staticmethod
    def forward(ctx, a, shape):
        ctx.shape = a.shape
        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.reshape_forward(a.data, shape)
        else:
            out_data = ops_gpu.reshape_forward(ctx.cl_ctx, ctx.cl_queue, a, shape)
        
        out = Tensor(out_data, requires_grad=requires_grad, 
                     is_leaf=is_leaf, device=a.device)
        
        assert np.prod(out.shape) == np.prod(a.shape), "Inconsistent array reshape size"

        out.children = [a]
        out.op = 'reshape'
        return out

    @staticmethod
    def backward(ctx, grad_output):
        if grad_output.device == Device.CPU:
            grad = ops_cpu.reshape_backward(grad_output.data, ctx.shape)
        else:
            grad = ops_gpu.reshape_backward(ctx.cl_ctx, ctx.cl_queue, 
                                            grad_output.data, ctx.shape)

        return Tensor(grad, device=grad_output.device), None


class Max(Function):
    @staticmethod
    def forward(ctx, a, axis=None):
        axis = [axis] if type(axis) == int else axis

        ctx.save_for_backward(a)

        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.max_forward(a.data, axis)
        else:
            out_data = ops_gpu.max_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis)
        
        ctx.axis, ctx.out = axis, out_data

        out = Tensor(out_data, requires_grad=requires_grad, 
                     is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'max'

        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        axis, out = ctx.axis, ctx.out
        inp = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad = ops_cpu.max_backward(grad_output, inp.data, out, axis)
        else:
            grad = ops_gpu.max_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, 
                                        inp.data, out, axis)
            
        return Tensor(grad, device=grad_output.device), None


class Min(Function):
    @staticmethod
    def forward(ctx, a, axis=None):
        axis = [axis] if type(axis) == int else axis

        ctx.save_for_backward(a)

        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.min_forward(a.data, axis)
        else:
            out_data = ops_gpu.min_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis)
        
        ctx.axis, ctx.out = axis, out_data

        out = Tensor(out_data, requires_grad=requires_grad, 
                     is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'min'

        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        axis, out = ctx.axis, ctx.out
        inp = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad = ops_cpu.min_backward(grad_output, inp.data, out, axis)
        else:
            grad = ops_gpu.min_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, 
                                        inp.data, out, axis)

        return Tensor(grad, device=grad_output.device), None


class Log(Function):
    @staticmethod
    def forward(ctx, a):
        ctx.save_for_backward(a)
        requires_grad = a.requires_grad
    
        if a.device == Device.CPU:
            out_data = ops_cpu.log_forward(a.data)
        else:
            out_data = ops_gpu.log_forward(ctx.cl_ctx, ctx.cl_queue, a.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                     is_leaf=not requires_grad, device=a.device)
        out.children = [a]   
        out.op = 'log'                
        return out

    @staticmethod
    def backward(ctx, grad_output):
        a = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad_a = ops_cpu.log_backward(grad_output.data, a.data)
        else:
            grad_a = ops_gpu.log_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data)

        return Tensor(grad_a, device=grad_output.device), None
    

class Exp(Function):
    @staticmethod
    def forward(ctx, a):   
        ctx.save_for_backward(a)
        requires_grad = a.requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.exp_forward(a.data)
        else:
            out_data = ops_gpu.exp_forward(ctx.cl_ctx, ctx.cl_queue, a.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                            is_leaf=not requires_grad, device=a.device)
        out.children = [a]
        out.op = 'exp'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        a = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad_a = ops_cpu.exp_backward(grad_output.data, a.data)
        else:
            grad_a = ops_gpu.exp_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data)

        return Tensor(grad_a, device=grad_output.device), None


class Add(Function):
    @staticmethod
    def forward(ctx, a, b):
        if a.device != b.device:
            if a.device == Device.CPU and b.device == Device.GPU:
                a.gpu()
            elif a.device == Device.GPU and b.device == Device.CPU:
                b.gpu()

        ctx.save_for_backward(a, b)
        requires_grad = a.requires_grad or b.requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.add_forward(a.data, b.data)
        else:
            out_data = ops_gpu.add_forward(ctx.cl_ctx, ctx.cl_queue, a.data, b.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                            is_leaf=not requires_grad, device=a.device)
        out.children = [a, b]
        out.op = 'add'
        return out

    @staticmethod
    def backward(ctx, grad_output):
        a, b = ctx.saved_tensors

        if grad_output.device == Device.CPU:
            grad_a, grad_b = ops_cpu.add_backward(grad_output.data, a.shape, b.shape)
        else:
            grad_a, grad_b = ops_gpu.add_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data,
                                                  a.shape, b.shape)
        return Tensor(grad_a, device=grad_output.device), Tensor(grad_b, device=grad_output.device)


class Sum(Function):
    @staticmethod
    def forward(ctx, a, axis):
        ctx.axis = axis
        ctx.save_for_backward(a)

        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.sum_forward(a.data, axis)
        else:
            out_data = ops_gpu.sum_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis)

        out = Tensor(out_data, requires_grad=requires_grad, 
                     is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'sum'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        axis = ctx.axis
        a = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad = ops_cpu.sum_backward(grad_output.data, a.data, axis)
        else:
            grad = ops_gpu.sum_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data, axis)
        
        return Tensor(grad, device=grad_output.device), None, None


class Neg(Function):
    @staticmethod
    def forward(ctx, a):
        requires_grad = a.requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.neg_forward(a.data)
        else:
            out_data = ops_gpu.neg_forward(ctx.cl_ctx, ctx.cl_queue, a.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                            is_leaf=not a.requires_grad, device=a.device)
        out.children = [a]
        out.op = 'neg'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        if grad_output.device == Device.CPU:
            grad = ops_cpu.neg_backward(grad_output.data)
        else:
            grad = ops_gpu.neg_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data)
        return Tensor(grad, device=grad_output.device), None


class MatMul(Function):
    @staticmethod
    def forward(ctx, a, b):
        if a.shape[1] != b.shape[0]:
            raise Exception(f"Shapes don't match: {a.shape}, {b.shape}")

        if a.device != b.device:
            if a.device == Device.CPU and b.device == Device.GPU:
                a.gpu()
            elif a.device == Device.GPU and b.device == Device.CPU:
                b.gpu()

        ctx.save_for_backward(a, b)
        requires_grad = a.requires_grad or b.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.matmul_forward(a.data, b.data)
        else:
            out_data = ops_gpu.matmul_forward(ctx.cl_ctx, ctx.cl_queue, a.data, b.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                            is_leaf=is_leaf, device=a.device)
        out.children = [a, b]
        out.op = 'matmul'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        a, b = ctx.saved_tensors

        if grad_output.device == Device.CPU:
            grad_a, grad_b = ops_cpu.matmul_backward(grad_output.data, a.data, b.data)
        else:
            grad_a, grad_b = ops_gpu.matmul_backward(ctx.cl_ctx, ctx.cl_queue, 
                                                     grad_output.data, a.data, b.data)
        return Tensor(grad_a, device=grad_output.device), Tensor(grad_b, device=grad_output.device)


class Pow(Function):
    @staticmethod
    def forward(ctx, a, exp):
        if not isinstance(exp, (int, float)):
            raise Exception("Power can only be float or int")

        ctx.save_for_backward(a)
        ctx.exp = exp

        requires_grad = a.requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.pow_forward(a.data, exp)
        else:
            out_data = ops_gpu.pow_forward(ctx.cl_ctx, ctx.cl_queue, a.data, exp)

        out = Tensor(out_data, requires_grad=requires_grad, 
                            is_leaf=not requires_grad, device=a.device)
        out.children = [a]
        out.op = 'pow'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        exp = ctx.exp
        a = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad_a = ops_cpu.pow_backward(grad_output.data, a.data, exp)
        else:
            grad_a = ops_gpu.pow_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data, exp)

        return Tensor(grad_a, device=grad_output.device), None


class Mul(Function):
    @staticmethod
    def forward(ctx, a, b):
        if a.device != b.device:
            if a.device == Device.CPU and b.device == Device.GPU:
                a.gpu()
            elif a.device == Device.GPU and b.device == Device.CPU:
                b.gpu()

        ctx.save_for_backward(a, b)
        requires_grad = a.requires_grad or b.requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.mul_forward(a.data, b.data)
        else:
            out_data = ops_gpu.mul_forward(ctx.cl_ctx, ctx.cl_queue, a.data, b.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                     is_leaf=not requires_grad, device=a.device)
        out.children = [a, b]
        out.op = 'mul'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        a, b = ctx.saved_tensors

        if grad_output.device == Device.CPU:
            grad_a, grad_b = ops_cpu.mul_backward(grad_output.data, a.data, b.data)
        else:
            grad_a, grad_b = ops_gpu.mul_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data, b.data)
        
        return Tensor(grad_a, device=grad_output.device), Tensor(grad_b, device=grad_output.device)


class ReLU(Function):
    @staticmethod
    def forward(ctx, a):
        ctx.save_for_backward(a)

        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.relu_forward(a.data)
        else:
            out_data = ops_gpu.relu_forward(ctx.cl_ctx, ctx.cl_queue, a.data)

        out = Tensor(out_data, requires_grad=requires_grad, 
                            is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'relu'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        a = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad_a = ops_cpu.relu_backward(grad_output.data, a.data)
        else:
            grad_a = ops_gpu.relu_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data)

        return Tensor(grad_a, device=grad_output.device), None


class Sigmoid(Function):
    @staticmethod
    def forward(ctx, a):
        ctx.save_for_backward(a)

        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.sigmoid_forward(a.data)
        else:
            out_data = ops_gpu.sigmoid_forward(ctx.cl_ctx, ctx.cl_queue, a.data)
        
        out = Tensor(out_data, requires_grad=requires_grad,
                            is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'sigmoid'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        a = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad_a = ops_cpu.sigmoid_backward(grad_output.data, a.data)
        else:
            grad_a = ops_gpu.sigmoid_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data)

        return Tensor(grad_a, device=grad_output.device), None


class Tanh(Function):
    @staticmethod
    def forward(ctx, a):
        ctx.save_for_backward(a)

        requires_grad = a.requires_grad
        is_leaf = not requires_grad

        if a.device == Device.CPU:
            out_data = ops_cpu.tanh_forward(a.data)
        else:
            out_data = ops_gpu.tanh_forward(ctx.cl_ctx, ctx.cl_queue, a.data)
        
        out = Tensor(out_data, requires_grad=requires_grad,
                            is_leaf=is_leaf, device=a.device)
        out.children = [a]
        out.op = 'tanh'
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        a = ctx.saved_tensors[0]

        if grad_output.device == Device.CPU:
            grad_a = ops_cpu.tanh_backward(grad_output.data, a.data)
        else:
            grad_a = ops_gpu.tanh_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, a.data)

        return Tensor(grad_a, device=grad_output.device), None


class Conv1d(Function):
    @staticmethod
    def forward(ctx, x, weight, stride):
        requires_grad = x.requires_grad
        is_leaf = not x.requires_grad

        if x.device == Device.CPU:
            out, x_reshaped = ops_cpu.conv1d_forward(x.data, weight.data, stride)
            ctx.x_reshaped = x_reshaped
        else:
            out = ops_gpu.conv1d_forward(ctx.cl_ctx, ctx.cl_queue, x.data, weight.data, stride)
        
        ctx.save_for_backward(x, weight)
        ctx.stride = stride

        out = Tensor(out, device=x.device, requires_grad=requires_grad, is_leaf=is_leaf)
        out.children = [x, weight]
        out.op = 'conv1d'
        return out

    @staticmethod
    def backward(ctx, grad_output):
        x, weight = ctx.saved_tensors
        stride = ctx.stride

        if grad_output.device == Device.CPU:
            x_reshaped = ctx.x_reshaped
            grad_x, grad_weight = ops_cpu.conv1d_backward(grad_output.data, x, x_reshaped, weight.data, stride)
        else:
            grad_x, grad_weight = ops_gpu.conv1d_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, 
                                                          x.data, weight.data, stride)

        return Tensor(grad_x, device=grad_output.device), Tensor(grad_weight, device=grad_output.device)
        

class Conv2d(Function):
    @staticmethod
    def forward(ctx, x, weight, stride):
        requires_grad = x.requires_grad
        is_leaf = not requires_grad

        if x.device == Device.CPU:
            out, x_reshaped = ops_cpu.conv2d_forward(x.data, weight.data, stride)
            ctx.x_reshaped = x_reshaped
        else:
            out = ops_gpu.conv2d_forward(ctx.cl_ctx, ctx.cl_queue, x.data, weight.data, stride)

        ctx.save_for_backward(x, weight)
        ctx.stride = stride

        out = Tensor(out, requires_grad=requires_grad, is_leaf=is_leaf, device=x.device)
        out.children = [x, weight]
        out.op = 'conv2d'
        
        return out
    
    @staticmethod
    def backward(ctx, grad_output):
        x, weight = ctx.saved_tensors
        stride = ctx.stride

        if grad_output.device == Device.CPU:
            x_reshaped = ctx.x_reshaped
            grad_x, grad_weight = ops_cpu.conv2d_backward(grad_output.data, x, x_reshaped, weight.data, stride)
        else:
            grad_x, grad_weight = ops_gpu.conv2d_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, 
                                                          x.data, weight.data, stride)

        grad_x = Tensor(grad_x, device=grad_output.device)
        grad_weight = Tensor(grad_weight, device=grad_output.device)

        return grad_x, grad_weight